@@ -936,7 +936,7 @@ def has_flag(compiler, flagname):
 def cpp_flag(compiler):
     """Return the -std=c++[11/14] compiler flag.
 
-    The c++14 is prefered over c++11 (when it is available).
+    The c++14 is preferred over c++11 (when it is available).
     """
     if has_flag(compiler, "-std=gnu++14"):
         return "-std=gnu++14"
