@@ -1,5 +1,3 @@
-from __future__ import division, with_statement
-
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
 __license__ = """
@@ -27,24 +25,32 @@ import numpy as np
 import re
 from pymbolic.parser import Parser as ExpressionParserBase
 from pymbolic.mapper import CombineMapper
-import pymbolic.primitives
+import pymbolic.primitives as p
 from pymbolic.mapper.c_code import CCodeMapper as CCodeMapperBase
+from sys import intern
 
 from warnings import warn
 
 import pytools.lex
-import re
-
-
 
 
 class TranslatorWarning(UserWarning):
     pass
 
+
 class TranslationError(RuntimeError):
     pass
 
 
+def complex_type_name(dtype):
+    if dtype == np.complex64:
+        return "cfloat"
+    if dtype == np.complex128:
+        return "cdouble"
+    else:
+        raise RuntimeError
+
+
 # {{{ AST components
 
 def dtype_to_ctype(dtype):
@@ -77,9 +83,7 @@ def dtype_to_ctype(dtype):
     elif dtype == np.complex128:
         return "cdouble_t"
     else:
-        raise ValueError, "unable to map dtype '%s'" % dtype
-
-
+        raise ValueError("unable to map dtype '%s'" % dtype)
 
 
 class POD(cgen.POD):
@@ -89,8 +93,6 @@ class POD(cgen.POD):
 # }}}
 
 
-
-
 # {{{ expression parser
 
 _less_than = intern("less_than")
@@ -104,7 +106,8 @@ _not = intern("not")
 _and = intern("and")
 _or = intern("or")
 
-class TypedLiteral(pymbolic.primitives.Leaf):
+
+class TypedLiteral(p.Leaf):
     def __init__(self, value, dtype):
         self.value = value
         self.dtype = np.dtype(dtype)
@@ -114,21 +117,34 @@ class TypedLiteral(pymbolic.primitives.Leaf):
 
     mapper_method = intern("map_literal")
 
+
+def simplify_typed_literal(expr):
+    if (isinstance(expr, p.Product)
+            and len(expr.children) == 2
+            and isinstance(expr.children[1], TypedLiteral)
+            and p.is_constant(expr.children[0])
+            and expr.children[0] == -1):
+        tl = expr.children[1]
+        return TypedLiteral("-"+tl.value, tl.dtype)
+    else:
+        return expr
+
+
 class FortranExpressionParser(ExpressionParserBase):
     # FIXME double/single prec literals
 
     lex_table = [
-            (_less_than, pytools.lex.RE(r"\.lt\.", re.I)),
-            (_greater_than, pytools.lex.RE(r"\.gt\.", re.I)),
-            (_less_equal, pytools.lex.RE(r"\.le\.", re.I)),
-            (_greater_equal, pytools.lex.RE(r"\.ge\.", re.I)),
-            (_equal, pytools.lex.RE(r"\.eq\.", re.I)),
-            (_not_equal, pytools.lex.RE(r"\.ne\.", re.I)),
+        (_less_than, pytools.lex.RE(r"\.lt\.", re.I)),
+        (_greater_than, pytools.lex.RE(r"\.gt\.", re.I)),
+        (_less_equal, pytools.lex.RE(r"\.le\.", re.I)),
+        (_greater_equal, pytools.lex.RE(r"\.ge\.", re.I)),
+        (_equal, pytools.lex.RE(r"\.eq\.", re.I)),
+        (_not_equal, pytools.lex.RE(r"\.ne\.", re.I)),
 
-            (_not, pytools.lex.RE(r"\.not\.", re.I)),
-            (_and, pytools.lex.RE(r"\.and\.", re.I)),
-            (_or, pytools.lex.RE(r"\.or\.", re.I)),
-            ] + ExpressionParserBase.lex_table
+        (_not, pytools.lex.RE(r"\.not\.", re.I)),
+        (_and, pytools.lex.RE(r"\.and\.", re.I)),
+        (_or, pytools.lex.RE(r"\.or\.", re.I)),
+        ] + ExpressionParserBase.lex_table
 
     def __init__(self, tree_walker):
         self.tree_walker = tree_walker
@@ -138,9 +154,8 @@ class FortranExpressionParser(ExpressionParserBase):
     def parse_terminal(self, pstate):
         scope = self.tree_walker.scope_stack[-1]
 
-        from pymbolic.primitives import Subscript, Call, Variable
         from pymbolic.parser import (
-                _identifier, _openpar, _closepar, _float)
+            _identifier, _openpar, _closepar, _float)
 
         next_tag = pstate.next_tag()
         if next_tag is _float:
@@ -154,11 +169,11 @@ class FortranExpressionParser(ExpressionParserBase):
             if value.startswith("."):
                 prev_value = value
                 value = "0"+value
-                print value, prev_value
+                print(value, prev_value)
             elif value.startswith("-."):
                 prev_value = value
                 value = "-0"+value[1:]
-                print value, prev_value
+                print(value, prev_value)
             return TypedLiteral(value, dtype)
 
         elif next_tag is _identifier:
@@ -168,17 +183,17 @@ class FortranExpressionParser(ExpressionParserBase):
                 # not a subscript
                 scope.use_name(name)
 
-                return Variable(name)
+                return p.Variable(name)
 
-            left_exp = Variable(name)
+            left_exp = p.Variable(name)
 
             pstate.advance()
             pstate.expect_not_end()
 
             if scope.is_known(name):
-                cls = Subscript
+                cls = p.Subscript
             else:
-                cls = Call
+                cls = p.Call
 
             if pstate.next_tag is _closepar:
                 pstate.advance()
@@ -223,14 +238,14 @@ class FortranExpressionParser(ExpressionParserBase):
                 _PREC_CALL, _PREC_COMPARISON, _openpar,
                 _PREC_LOGICAL_OR, _PREC_LOGICAL_AND)
         from pymbolic.primitives import (
-                ComparisonOperator, LogicalAnd, LogicalOr)
+                Comparison, LogicalAnd, LogicalOr)
 
         next_tag = pstate.next_tag()
         if next_tag is _openpar and _PREC_CALL > min_precedence:
             raise TranslationError("parenthesis operator only works on names")
         elif next_tag in self.COMP_MAP and _PREC_COMPARISON > min_precedence:
             pstate.advance()
-            left_exp = ComparisonOperator(
+            left_exp = Comparison(
                     left_exp,
                     self.COMP_MAP[next_tag],
                     self.parse_expression(pstate, _PREC_COMPARISON))
@@ -254,7 +269,10 @@ class FortranExpressionParser(ExpressionParserBase):
                 assert len(left_exp) == 2
                 r, i = left_exp
 
-                dtype = (r.dtype.type(0) + i.dtype.type(0))
+                r = simplify_typed_literal(r)
+                i = simplify_typed_literal(i)
+
+                dtype = (r.dtype.type(0) + i.dtype.type(0)).dtype
                 if dtype == np.float32:
                     dtype = np.complex64
                 else:
@@ -266,6 +284,7 @@ class FortranExpressionParser(ExpressionParserBase):
 
 # }}}
 
+
 # {{{ expression generator
 
 class TypeInferenceMapper(CombineMapper):
@@ -298,45 +317,57 @@ class TypeInferenceMapper(CombineMapper):
             else:
                 raise RuntimeError("unexpected complex type")
 
+        elif name in ["imag", "real", "abs", "dble"]:
+            arg, = expr.parameters
+            base_dtype = self.rec(arg)
+
+            if base_dtype == np.complex128:
+                return np.dtype(np.float64)
+            elif base_dtype == np.complex64:
+                return np.dtype(np.float32)
+            else:
+                return base_dtype
+
         else:
             return CombineMapper.map_call(self, expr)
 
 
-
-
 class ComplexCCodeMapper(CCodeMapperBase):
     def __init__(self, infer_type):
         CCodeMapperBase.__init__(self)
         self.infer_type = infer_type
 
-    def complex_type_name(self, dtype):
-        if dtype == np.complex64:
-            return "cfloat"
-        if dtype == np.complex128:
-            return "cdouble"
-        else:
-            raise RuntimeError
-
     def map_sum(self, expr, enclosing_prec):
         tgt_dtype = self.infer_type(expr)
-        is_complex = tgt_dtype.kind == 'c'
+        is_complex = tgt_dtype.kind == "c"
 
         if not is_complex:
             return CCodeMapperBase.map_sum(self, expr, enclosing_prec)
         else:
-            tgt_name = self.complex_type_name(tgt_dtype)
+            tgt_name = complex_type_name(tgt_dtype)
 
             reals = [child for child in expr.children
-                    if 'c' != self.infer_type(child).kind]
+                    if "c" != self.infer_type(child).kind]
             complexes = [child for child in expr.children
-                    if 'c' == self.infer_type(child).kind]
+                    if "c" == self.infer_type(child).kind]
 
-            from pymbolic.mapper.stringifier import PREC_SUM
+            from pymbolic.mapper.stringifier import PREC_SUM, PREC_NONE
             real_sum = self.join_rec(" + ", reals, PREC_SUM)
-            complex_sum = self.join_rec(" + ", complexes, PREC_SUM)
+
+            if len(complexes) == 1:
+                myprec = PREC_SUM
+            else:
+                myprec = PREC_NONE
+
+            complex_sum = self.rec(complexes[0], myprec)
+            for child in complexes[1:]:
+                complex_sum = "{}_add({}, {})".format(
+                        tgt_name, complex_sum,
+                        self.rec(child, PREC_NONE))
 
             if real_sum:
-                result = "%s_fromreal(%s) + %s" % (tgt_name, real_sum, complex_sum)
+                result = "{}_add({}_fromreal({}), {})".format(
+                        tgt_name, tgt_name, real_sum, complex_sum)
             else:
                 result = complex_sum
 
@@ -344,17 +375,17 @@ class ComplexCCodeMapper(CCodeMapperBase):
 
     def map_product(self, expr, enclosing_prec):
         tgt_dtype = self.infer_type(expr)
-        is_complex = 'c' == tgt_dtype.kind
+        is_complex = "c" == tgt_dtype.kind
 
         if not is_complex:
             return CCodeMapperBase.map_product(self, expr, enclosing_prec)
         else:
-            tgt_name = self.complex_type_name(tgt_dtype)
+            tgt_name = complex_type_name(tgt_dtype)
 
             reals = [child for child in expr.children
-                    if 'c' != self.infer_type(child).kind]
+                    if "c" != self.infer_type(child).kind]
             complexes = [child for child in expr.children
-                    if 'c' == self.infer_type(child).kind]
+                    if "c" == self.infer_type(child).kind]
 
             from pymbolic.mapper.stringifier import PREC_PRODUCT, PREC_NONE
             real_prd = self.join_rec("*", reals, PREC_PRODUCT)
@@ -366,13 +397,12 @@ class ComplexCCodeMapper(CCodeMapperBase):
 
             complex_prd = self.rec(complexes[0], myprec)
             for child in complexes[1:]:
-                complex_prd = "%s_mul(%s, %s)" % (
+                complex_prd = "{}_mul({}, {})".format(
                         tgt_name, complex_prd,
                         self.rec(child, PREC_NONE))
 
             if real_prd:
-                # elementwise semantics are correct
-                result = "%s * %s" % (real_prd, complex_prd)
+                result = f"{tgt_name}_rmul({real_prd}, {complex_prd})"
             else:
                 result = complex_prd
 
@@ -380,30 +410,32 @@ class ComplexCCodeMapper(CCodeMapperBase):
 
     def map_quotient(self, expr, enclosing_prec):
         from pymbolic.mapper.stringifier import PREC_NONE
-        n_complex = 'c' == self.infer_type(expr.numerator).kind
-        d_complex = 'c' == self.infer_type(expr.denominator).kind
+        n_complex = "c" == self.infer_type(expr.numerator).kind
+        d_complex = "c" == self.infer_type(expr.denominator).kind
 
         tgt_dtype = self.infer_type(expr)
 
         if not (n_complex or d_complex):
             return CCodeMapperBase.map_quotient(self, expr, enclosing_prec)
         elif n_complex and not d_complex:
-            # elementwise semnatics are correct
-            return CCodeMapperBase.map_quotient(self, expr, enclosing_prec)
+            return "{}_divider({}, {})".format(
+                    complex_type_name(tgt_dtype),
+                    self.rec(expr.numerator, PREC_NONE),
+                    self.rec(expr.denominator, PREC_NONE))
         elif not n_complex and d_complex:
-            return "%s_rdivide(%s, %s)" % (
-                    self.complex_type_name(tgt_dtype),
+            return "{}_rdivide({}, {})".format(
+                    complex_type_name(tgt_dtype),
                     self.rec(expr.numerator, PREC_NONE),
                     self.rec(expr.denominator, PREC_NONE))
         else:
-            return "%s_divide(%s, %s)" % (
-                    self.complex_type_name(tgt_dtype),
+            return "{}_divide({}, {})".format(
+                    complex_type_name(tgt_dtype),
                     self.rec(expr.numerator, PREC_NONE),
                     self.rec(expr.denominator, PREC_NONE))
 
     def map_remainder(self, expr, enclosing_prec):
         tgt_dtype = self.infer_type(expr)
-        if 'c' == tgt_dtype.kind:
+        if "c" == tgt_dtype.kind:
             raise RuntimeError("complex remainder not defined")
 
         return CCodeMapperBase.map_remainder(self, expr, enclosing_prec)
@@ -412,32 +444,30 @@ class ComplexCCodeMapper(CCodeMapperBase):
         from pymbolic.mapper.stringifier import PREC_NONE
 
         tgt_dtype = self.infer_type(expr)
-        if 'c' == tgt_dtype.kind:
+        if "c" == tgt_dtype.kind:
             if expr.exponent in [2, 3, 4]:
                 value = expr.base
                 for i in range(expr.exponent-1):
                     value = value * expr.base
                 return self.rec(value, enclosing_prec)
             else:
-                b_complex = 'c' == self.infer_type(expr.base).kind
-                e_complex = 'c' == self.infer_type(expr.exponent).kind
+                b_complex = "c" == self.infer_type(expr.base).kind
+                e_complex = "c" == self.infer_type(expr.exponent).kind
 
                 if b_complex and not e_complex:
-                    return "%s_powr(%s, %s)" % (
-                            self.complex_type_name(tgt_dtype),
+                    return "{}_powr({}, {})".format(
+                            complex_type_name(tgt_dtype),
                             self.rec(expr.base, PREC_NONE),
                             self.rec(expr.exponent, PREC_NONE))
                 else:
-                    return "%s_pow(%s, %s)" % (
-                            self.complex_type_name(tgt_dtype),
+                    return "{}_pow({}, {})".format(
+                            complex_type_name(tgt_dtype),
                             self.rec(expr.base, PREC_NONE),
                             self.rec(expr.exponent, PREC_NONE))
 
         return CCodeMapperBase.map_power(self, expr, enclosing_prec)
 
 
-
-
 class CCodeMapper(ComplexCCodeMapper):
     # Whatever is needed to mop up after Fortran goes here.
     # Stuff that deals with generating real-valued code
@@ -450,7 +480,7 @@ class CCodeMapper(ComplexCCodeMapper):
 
     def map_subscript(self, expr, enclosing_prec):
         idx_dtype = self.infer_type(expr.index)
-        if not 'i' == idx_dtype.kind or 'u' == idx_dtype.kind:
+        if not "i" == idx_dtype.kind or "u" == idx_dtype.kind:
             ind_prefix = "(int) "
         else:
             ind_prefix = ""
@@ -471,28 +501,42 @@ class CCodeMapper(ComplexCCodeMapper):
         from pymbolic.mapper.stringifier import PREC_NONE
 
         tgt_dtype = self.infer_type(expr)
+        arg_dtypes = [self.infer_type(par) for par in expr.parameters]
 
         name = expr.function.name
-        if 'f' == tgt_dtype.kind and name == "abs":
+        if "f" == tgt_dtype.kind and name == "abs":
             name = "fabs"
 
-        if 'c' == tgt_dtype.kind:
+        elif "c" == tgt_dtype.kind:
             if name in ["conjg", "dconjg"]:
                 name = "conj"
 
             if name[:2] == "cd" and name[2:] in ["log", "exp", "sqrt"]:
                 name = name[2:]
 
-            if name == "aimag":
-                name = "imag"
-
             if name == "dble":
                 name = "real"
 
-            name = "%s_%s" % (
-                    self.complex_type_name(tgt_dtype),
+            name = "{}_{}".format(
+                    complex_type_name(tgt_dtype),
                     name)
 
+        elif name in ["aimag", "real", "imag"] and tgt_dtype.kind == "f":
+            arg_dtype, = arg_dtypes
+
+            if name == "aimag":
+                name = "imag"
+
+            name = "{}_{}".format(
+                    complex_type_name(arg_dtype),
+                    name)
+
+        elif "c" == tgt_dtype.kind and name == "abs":
+            arg_dtype, = arg_dtypes
+
+            name = "%s_abs" % (
+                    complex_type_name(arg_dtype))
+
         return self.format("%s(%s)",
                 name,
                 self.join_rec(", ", expr.parameters, PREC_NONE))
@@ -519,7 +563,10 @@ class CCodeMapper(ComplexCCodeMapper):
         from pymbolic.mapper.stringifier import PREC_NONE
         if expr.dtype.kind == "c":
             r, i = expr.value
-            return "{ %s, %s }" % (self.rec(r, PREC_NONE), self.rec(i, PREC_NONE))
+            return "{}_new({}, {})".format(
+                    complex_type_name(expr.dtype),
+                    self.rec(r, PREC_NONE),
+                    self.rec(i, PREC_NONE))
         else:
             return expr.value
 
@@ -529,7 +576,7 @@ class CCodeMapper(ComplexCCodeMapper):
 
 # }}}
 
-class Scope(object):
+class Scope:
     def __init__(self, subprogram_name, arg_names=set()):
         self.subprogram_name = subprogram_name
 
@@ -556,8 +603,8 @@ class Scope(object):
 
     def known_names(self):
         return (self.used_names
-                | set(self.dim_map.iterkeys())
-                | set(self.type_map.iterkeys()))
+                | set(self.dim_map.keys())
+                | set(self.type_map.keys()))
 
     def is_known(self, name):
         return (name in self.used_names
@@ -591,14 +638,12 @@ class Scope(object):
     def translate_var_name(self, name):
         shape = self.dim_map.get(name)
         if name in self.data and shape is not None:
-            return "%s_%s" % (self.subprogram_name, name)
+            return f"{self.subprogram_name}_{name}"
         else:
             return name
 
 
-
-
-class FTreeWalkerBase(object):
+class FTreeWalkerBase:
     def __init__(self):
         self.scope_stack = []
 
@@ -625,7 +670,7 @@ class FTreeWalkerBase(object):
 
     ENTITY_RE = re.compile(
             r"^(?P<name>[_0-9a-zA-Z]+)"
-            "(\((?P<shape>[-+*0-9:a-zA-Z,]+)\))?$")
+            r"(\((?P<shape>[-+*0-9:a-zA-Z,]+)\))?$")
 
     def parse_dimension_specs(self, dim_decls):
         def parse_bounds(bounds_str):
@@ -633,7 +678,7 @@ class FTreeWalkerBase(object):
 
             assert 1 <= len(start_end) <= 2
 
-            return (self.parse_expr(s) for s in start_end)
+            return tuple(self.parse_expr(s) for s in start_end)
 
         for decl in dim_decls:
             entity_match = self.ENTITY_RE.match(decl)
@@ -661,13 +706,12 @@ class FTreeWalkerBase(object):
     # }}}
 
 
-
 class ArgumentAnalayzer(FTreeWalkerBase):
     def __init__(self):
         FTreeWalkerBase.__init__(self)
 
         # map (func, arg_nr) to
-        # 'w' for 'needs pointer'
+        # "w" for "needs pointer"
         # [] for no obstacle to de-pointerification known
         # [(func_name, arg_nr), ...] # depends on how this arg is used
 
@@ -768,11 +812,9 @@ class ArgumentAnalayzer(FTreeWalkerBase):
 
         lhs = self.parse_expr(node.variable)
 
-
-        from pymbolic.primitives import Subscript, Call
-        if isinstance(lhs, Subscript):
+        if isinstance(lhs, p.Subscript):
             lhs_name = lhs.aggregate.name
-        elif isinstance(lhs, Call):
+        elif isinstance(lhs, p.Call):
             # in absence of dim info, subscripts get parsed as calls
             lhs_name = lhs.function.name
         else:
@@ -808,11 +850,10 @@ class ArgumentAnalayzer(FTreeWalkerBase):
     def map_Call(self, node):
         scope = self.scope_stack[-1]
 
-        from pymbolic.primitives import Subscript, Variable
         for i, arg_str in enumerate(node.items):
             arg = self.parse_expr(arg_str)
-            if isinstance(arg, (Variable, Subscript)):
-                if isinstance(arg, Subscript):
+            if isinstance(arg, (p.Variable, p.Subscript)):
+                if isinstance(arg, p.Subscript):
                     arg_name = arg.aggregate.name
                 else:
                     arg_name = arg.name
@@ -869,9 +910,6 @@ class ArgumentAnalayzer(FTreeWalkerBase):
     # }}}
 
 
-
-
-
 # {{{ translator
 
 class F2CLTranslator(FTreeWalkerBase):
@@ -892,7 +930,6 @@ class F2CLTranslator(FTreeWalkerBase):
         scope = self.scope_stack[-1]
         return POD(scope.get_type(name), name)
 
-
     def get_declarations(self):
         scope = self.scope_stack[-1]
 
@@ -907,12 +944,12 @@ class F2CLTranslator(FTreeWalkerBase):
 
             if shape is not None:
                 dim_stmt = cgen.Statement(
-                    "dimension \"fortran\" %s[%s]" % (
+                    'dimension \"fortran\" {}[{}]'.format(
                         scope.translate_var_name(name),
                         ", ".join(gen_shape(s) for s in shape)
                         ))
 
-                # cannot omit 'dimension' decl even for rank-1 args:
+                # cannot omit "dimension" decl even for rank-1 args:
                 result.append(dim_stmt)
 
             if name in scope.data:
@@ -933,7 +970,7 @@ class F2CLTranslator(FTreeWalkerBase):
                             cgen.Initializer(
                                 CLConstant(
                                     cgen.ArrayOf(self.get_declarator(
-                                        "%s_%s" % (scope.subprogram_name, name)))),
+                                        f"{scope.subprogram_name}_{name}"))),
                                 "{ %s }" % ",\n".join(self.gen_expr(x) for x in data)
                                 ))
             else:
@@ -941,9 +978,9 @@ class F2CLTranslator(FTreeWalkerBase):
                     if shape is not None:
                         result.append(cgen.Statement(
                             "%s %s[nitemsof(%s)]"
-                                % (
-                                    dtype_to_ctype(scope.get_type(name)),
-                                    name, name)))
+                            % (
+                                dtype_to_ctype(scope.get_type(name)),
+                                name, name)))
                     else:
                         result.append(self.get_declarator(name))
 
@@ -1003,8 +1040,7 @@ class F2CLTranslator(FTreeWalkerBase):
 
             return decl
 
-
-        result =  cgen.FunctionBody(
+        result = cgen.FunctionBody(
                 cgen.FunctionDeclaration(
                     cgen.Value("void", node.name),
                     [get_arg_decl(i, arg) for i, arg in enumerate(node.args)]
@@ -1051,7 +1087,7 @@ class F2CLTranslator(FTreeWalkerBase):
 
             ("integer", ""): np.int32,
             ("integer", "4"): np.int32,
-            ("complex", "8"): np.int64,
+            ("integer", "8"): np.int64,
             }
 
     def dtype_from_stmt(self, stmt):
@@ -1145,11 +1181,11 @@ class F2CLTranslator(FTreeWalkerBase):
         rhs_dtype = infer_type(rhs)
 
         # check for silent truncation of complex
-        if lhs_dtype.kind != 'c' and rhs_dtype.kind == 'c':
+        if lhs_dtype.kind != "c" and rhs_dtype.kind == "c":
             from pymbolic import var
             rhs = var("real")(rhs)
         # check for silent widening of real
-        if lhs_dtype.kind == 'c' and rhs_dtype.kind != 'c':
+        if lhs_dtype.kind == "c" and rhs_dtype.kind != "c":
             from pymbolic import var
             rhs = var("fromreal")(rhs)
 
@@ -1190,13 +1226,13 @@ class F2CLTranslator(FTreeWalkerBase):
             cast = self.force_casts.get(
                     (node.designator, i))
             if cast is not None:
-                result = "(%s) (%s)" % (cast, result)
+                result = f"({cast}) ({result})"
 
             return result
 
-        return cgen.Statement("%s(%s)" % (
+        return cgen.Statement("{}({})".format(
             node.designator,
-            ", ".join(transform_arg(i, arg_str) 
+            ", ".join(transform_arg(i, arg_str)
                 for i, arg_str in enumerate(node.items))))
 
     def map_Return(self, node):
@@ -1277,8 +1313,9 @@ class F2CLTranslator(FTreeWalkerBase):
                         % node.loopcontrol)
 
             if not isinstance(step, int):
-                print type(step)
-                raise TranslationError("non-constant steps not yet supported: %s" % step)
+                print(type(step))
+                raise TranslationError(
+                        "non-constant steps not yet supported: %s" % step)
 
             if step < 0:
                 comp_op = ">="
@@ -1286,9 +1323,9 @@ class F2CLTranslator(FTreeWalkerBase):
                 comp_op = "<="
 
             return cgen.For(
-                    "%s = %s" % (loop_var, self.gen_expr(start)),
-                    "%s %s %s" % (loop_var, comp_op, self.gen_expr(stop)),
-                    "%s += %s" % (loop_var, self.gen_expr(step)),
+                    "{} = {}".format(loop_var, self.gen_expr(start)),
+                    "{} {} {}".format(loop_var, comp_op, self.gen_expr(stop)),
+                    "{} += {}".format(loop_var, self.gen_expr(step)),
                     cgen.block_if_necessary(body))
 
         else:
@@ -1327,10 +1364,6 @@ class F2CLTranslator(FTreeWalkerBase):
 # }}}
 
 
-
-
-
-
 def f2cl(source, free_form=False, strict=True,
         addr_space_hints={}, force_casts={},
         do_arg_analysis=True,
@@ -1363,22 +1396,27 @@ def f2cl(source, free_form=False, strict=True,
         import pyopencl as cl
         ctx = cl.create_some_context()
         cl.Program(ctx, """
+            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
             #pragma OPENCL EXTENSION cl_khr_fp64: enable
+            #endif
             #include <pyopencl-complex.h>
             """).build()
     return str_mod
 
 
-
-
 def f2cl_files(source_file, target_file, **kwargs):
     mod = f2cl(open(source_file).read(), **kwargs)
     open(target_file, "w").write(mod)
 
 
-
-
 if __name__ == "__main__":
+    import logging
+    console = logging.StreamHandler()
+    console.setLevel(logging.DEBUG)
+    formatter = logging.Formatter("%(name)-12s: %(levelname)-8s %(message)s")
+    console.setFormatter(formatter)
+    logging.getLogger("fparser").addHandler(console)
+
     from cgen.opencl import CLConstant
 
     if 0:
