@@ -479,6 +479,11 @@ class Program:
 
     @classmethod
     def _process_build_options(cls, context, options, _add_include_path=False):
+        if options is None:
+            options = []
+        if isinstance(options, tuple):
+            options = list(options)
+
         options = _split_options_if_necessary(options)
 
         options = (options
@@ -496,7 +501,7 @@ class Program:
                 _options_to_bytestring(options),
                 _find_include_path(options))
 
-    def build(self, options=[], devices=None, cache_dir=None):
+    def build(self, options=None, devices=None, cache_dir=None):
         options_bytes, include_path = self._process_build_options(
                 self._context, options)
 
@@ -580,10 +585,14 @@ class Program:
 
     # }}}
 
-    def compile(self, options=[], devices=None, headers=[]):
+    def compile(self, options=None, devices=None, headers=None):
+        if headers is None:
+            headers = []
+
         options_bytes, _ = self._process_build_options(self._context, options)
 
-        self._get_prg().compile(options_bytes, devices, headers)
+        self._get_prg().compile(options_bytes, devices,
+                [(name, prg._get_prg()) for name, prg in headers])
         return self
 
     def __eq__(self, other):
@@ -1369,7 +1378,7 @@ def _add_functionality():
 
     for cls, (info_method, info_class, cacheable_attrs) \
             in cls_to_info_cls.items():
-        for info_name, info_value in info_class.__dict__.items():
+        for info_name, _info_value in info_class.__dict__.items():
             if info_name == "to_string" or info_name.startswith("_"):
                 continue
 
@@ -1409,7 +1418,7 @@ def create_some_context(interactive=None, answers=None):
 
         if "PYOPENCL_TEST" in os.environ:
             from pyopencl.tools import get_test_platforms_and_devices
-            for plat, devs in get_test_platforms_and_devices():
+            for _plat, devs in get_test_platforms_and_devices():
                 for dev in devs:
                     return Context([dev])
 
@@ -1471,7 +1480,7 @@ def create_some_context(interactive=None, answers=None):
 
             if platform is None:
                 answer = answer.lower()
-                for i, pf in enumerate(platforms):
+                for pf in platforms:
                     if answer in pf.name.lower():
                         platform = pf
                 if platform is None:
@@ -1493,7 +1502,7 @@ def create_some_context(interactive=None, answers=None):
                 return devices[int_choice]
 
         choice = choice.lower()
-        for i, dev in enumerate(devices):
+        for dev in devices:
             if choice in dev.name.lower():
                 return dev
         raise RuntimeError("input did not match any device")
@@ -1625,7 +1634,7 @@ def enqueue_copy(queue, dest, src, **kwargs):
         The size of the transfer is controlled by the size of the
         of the host-side buffer. If the host-side buffer
         is a :class:`numpy.ndarray`, you can control the transfer size by
-        transfering into a smaller 'view' of the target array, like this::
+        transferring into a smaller 'view' of the target array, like this::
 
             cl.enqueue_copy(queue, large_dest_numpy_array[:15], src_buffer)
 
@@ -1951,7 +1960,7 @@ def enqueue_svm_memfill(queue, dest, pattern, byte_count=None, wait_for=None):
 def enqueue_svm_migratemem(queue, svms, flags, wait_for=None):
     """
     :arg svms: a collection of Python buffer objects (e.g. :mod:`numpy`
-        arrrays), optionally wrapped in :class:`SVM` objects.
+        arrays), optionally wrapped in :class:`SVM` objects.
     :arg flags: a combination of :class:`mem_migration_flags`
 
     |std-enqueue-blurb|
