@@ -38,7 +38,7 @@ from pyopencl.tools import (dtype_to_ctype, VectorArg, ScalarArg,
 # {{{ elementwise kernel code generator
 
 def get_elwise_program(context, arguments, operation,
-        name="elwise_kernel", options=[],
+        name="elwise_kernel", options=None,
         preamble="", loop_prep="", after_loop="",
         use_range=False):
 
@@ -108,7 +108,7 @@ def get_elwise_program(context, arguments, operation,
 
 
 def get_elwise_kernel_and_types(context, arguments, operation,
-        name="elwise_kernel", options=[], preamble="", use_range=False,
+        name="elwise_kernel", options=None, preamble="", use_range=False,
         **kwargs):
 
     from pyopencl.tools import parse_arg_list, get_arg_offset_adjuster_code
@@ -165,7 +165,7 @@ def get_elwise_kernel_and_types(context, arguments, operation,
 
 
 def get_elwise_kernel(context, arguments, operation,
-        name="elwise_kernel", options=[], **kwargs):
+        name="elwise_kernel", options=None, **kwargs):
     """Return a L{pyopencl.Kernel} that performs the same scalar operation
     on one or several vectors.
     """
@@ -204,7 +204,7 @@ class ElementwiseKernel:
     """
 
     def __init__(self, context, arguments, operation,
-            name="elwise_kernel", options=[], **kwargs):
+            name="elwise_kernel", options=None, **kwargs):
         self.context = context
         self.arguments = arguments
         self.operation = operation
@@ -335,7 +335,7 @@ class ElementwiseTemplate(KernelTemplateBase):
 
     def build_inner(self, context, type_aliases=(), var_values=(),
             more_preamble="", more_arguments=(), declare_types=(),
-            options=()):
+            options=None):
         renderer = self.get_renderer(
                 type_aliases, var_values, context, options)
 
@@ -346,7 +346,7 @@ class ElementwiseTemplate(KernelTemplateBase):
 
         return ElementwiseKernel(context,
             arg_list, renderer(self.operation),
-            name=renderer(self.name), options=list(options),
+            name=renderer(self.name), options=options,
             preamble=(
                 type_decl_preamble
                 + "\n"
@@ -493,32 +493,36 @@ def real_dtype(dtype):
 
 
 @context_dependent_memoize
-def get_axpbyz_kernel(context, dtype_x, dtype_y, dtype_z):
+def get_axpbyz_kernel(context, dtype_x, dtype_y, dtype_z,
+                      x_is_scalar=False, y_is_scalar=False):
     result_t = dtype_to_ctype(dtype_z)
 
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
 
+    x = "x[0]" if x_is_scalar else "x[i]"
+    y = "y[0]" if y_is_scalar else "y[i]"
+
     if dtype_z.kind == "c":
         # a and b will always be complex here.
         z_ct = complex_dtype_to_name(dtype_z)
 
         if x_is_complex:
-            ax = f"{z_ct}_mul(a, {z_ct}_cast(x[i]))"
+            ax = f"{z_ct}_mul(a, {z_ct}_cast({x}))"
         else:
-            ax = f"{z_ct}_mulr(a, x[i])"
+            ax = f"{z_ct}_mulr(a, {x})"
 
         if y_is_complex:
-            by = f"{z_ct}_mul(b, {z_ct}_cast(y[i]))"
+            by = f"{z_ct}_mul(b, {z_ct}_cast({y}))"
         else:
-            by = f"{z_ct}_mulr(b, y[i])"
+            by = f"{z_ct}_mulr(b, {y})"
 
         result = f"{z_ct}_add({ax}, {by})"
     else:
         # real-only
 
-        ax = f"a*(({result_t}) x[i])"
-        by = f"b*(({result_t}) y[i])"
+        ax = f"a*(({result_t}) {x})"
+        by = f"b*(({result_t}) {y})"
 
         result = f"{ax} + {by}"
 
@@ -594,12 +598,13 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
 
 
 @context_dependent_memoize
-def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z):
+def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z,
+                        x_is_scalar=False, y_is_scalar=False):
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
 
-    x = "x[i]"
-    y = "y[i]"
+    x = "x[0]" if x_is_scalar else "x[i]"
+    y = "y[0]" if y_is_scalar else "y[i]"
 
     if x_is_complex and dtype_x != dtype_z:
         x = "{}_cast({})".format(complex_dtype_to_name(dtype_z), x)
@@ -626,13 +631,14 @@ def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z):
 
 
 @context_dependent_memoize
-def get_divide_kernel(context, dtype_x, dtype_y, dtype_z):
+def get_divide_kernel(context, dtype_x, dtype_y, dtype_z,
+                      x_is_scalar=False, y_is_scalar=False):
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
     z_is_complex = dtype_z.kind == "c"
 
-    x = "x[i]"
-    y = "y[i]"
+    x = "x[0]" if x_is_scalar else "x[i]"
+    y = "y[0]" if y_is_scalar else "y[i]"
 
     if z_is_complex and dtype_x != dtype_y:
         if x_is_complex and dtype_x != dtype_z:
@@ -809,13 +815,16 @@ def get_array_scalar_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b
 
 
 @context_dependent_memoize
-def get_array_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b):
+def get_array_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b,
+                           a_is_scalar=False, b_is_scalar=False):
+    a = "a[0]" if a_is_scalar else "a[i]"
+    b = "b[0]" if b_is_scalar else "b[i]"
     return get_elwise_kernel(context, [
         VectorArg(dtype_res, "out", with_offset=True),
         VectorArg(dtype_a, "a", with_offset=True),
         VectorArg(dtype_b, "b", with_offset=True),
         ],
-        "out[i] = a[i] %s b[i]" % operator,
+        f"out[i] = {a} {operator} {b}",
         name="binop_kernel")
 
 
@@ -1026,14 +1035,32 @@ def get_diff_kernel(context, dtype):
 
 
 @context_dependent_memoize
-def get_if_positive_kernel(context, crit_dtype, dtype):
+def get_if_positive_kernel(
+        context, crit_dtype, then_else_dtype,
+        is_then_array, is_else_array,
+        is_then_scalar, is_else_scalar):
+    if is_then_array:
+        then_ = "then_[0]" if is_then_scalar else "then_[i]"
+        then_arg = VectorArg(then_else_dtype, "then_", with_offset=True)
+    else:
+        assert is_then_scalar
+        then_ = "then_"
+        then_arg = ScalarArg(then_else_dtype, "then_")
+
+    if is_else_array:
+        else_ = "else_[0]" if is_else_scalar else "else_[i]"
+        else_arg = VectorArg(then_else_dtype, "else_", with_offset=True)
+    else:
+        assert is_else_scalar
+        else_ = "else_"
+        else_arg = ScalarArg(then_else_dtype, "else_")
+
     return get_elwise_kernel(context, [
-            VectorArg(dtype, "result", with_offset=True),
+            VectorArg(then_else_dtype, "result", with_offset=True),
             VectorArg(crit_dtype, "crit", with_offset=True),
-            VectorArg(dtype, "then_", with_offset=True),
-            VectorArg(dtype, "else_", with_offset=True),
+            then_arg, else_arg,
             ],
-            "result[i] = crit[i] > 0 ? then_[i] : else_[i]",
+            f"result[i] = crit[i] > 0 ? {then_} : {else_}",
             name="if_positive")
 
 # }}}
