@@ -1,6 +1,5 @@
 """Computation of reductions on vectors."""
 
-from __future__ import division
 
 __copyright__ = "Copyright (C) 2010 Andreas Kloeckner"
 
@@ -41,13 +40,15 @@ import numpy as np
 
 # {{{ kernel source
 
-KERNEL = """//CL//
-    #define GROUP_SIZE ${group_size}
-    #define READ_AND_MAP(i) (${map_expr})
-    #define REDUCE(a, b) (${reduce_expr})
+KERNEL = r"""//CL//
+    #define PCL_GROUP_SIZE ${group_size}
+    #define PCL_READ_AND_MAP(i) (${map_expr})
+    #define PCL_REDUCE(a, b) (${reduce_expr})
 
     % if double_support:
+        #if __OPENCL_C_VERSION__ < 120
         #pragma OPENCL EXTENSION cl_khr_fp64: enable
+        #endif
         #define PYOPENCL_DEFINE_CDOUBLE
     % endif
 
@@ -55,37 +56,43 @@ KERNEL = """//CL//
 
     ${preamble}
 
-    typedef ${out_type} out_type;
+    typedef ${out_type} pcl_out_type;
 
     __kernel void ${name}(
-      __global out_type *out, ${arguments},
-      unsigned int seq_count, unsigned int n)
+      __global pcl_out_type *pcl_out__base, long pcl_out__offset,
+      ${arguments}
+      long pcl_start, long pcl_step, long pcl_stop,
+      unsigned int pcl_seq_count, long n)
     {
-       ${arg_prep}
+        __global pcl_out_type *pcl_out = (__global pcl_out_type *) (
+            (__global char *) pcl_out__base + pcl_out__offset);
+        ${arg_prep}
 
-        __local out_type ldata[GROUP_SIZE];
+        __local pcl_out_type pcl_ldata[PCL_GROUP_SIZE];
 
-        unsigned int lid = get_local_id(0);
+        unsigned int pcl_lid = get_local_id(0);
 
-        unsigned int i = get_group_id(0)*GROUP_SIZE*seq_count + lid;
+        const long pcl_base_idx =
+            get_group_id(0)*PCL_GROUP_SIZE*pcl_seq_count + pcl_lid;
+        long i = pcl_start + pcl_base_idx * pcl_step;
 
-        out_type acc = ${neutral};
-        for (unsigned s = 0; s < seq_count; ++s)
+        pcl_out_type pcl_acc = ${neutral};
+        for (unsigned pcl_s = 0; pcl_s < pcl_seq_count; ++pcl_s)
         {
-          if (i >= n)
+          if (i >= pcl_stop)
             break;
-          acc = REDUCE(acc, READ_AND_MAP(i));
+          pcl_acc = PCL_REDUCE(pcl_acc, PCL_READ_AND_MAP(i));
 
-          i += GROUP_SIZE;
+          i += PCL_GROUP_SIZE*pcl_step;
         }
 
-        ldata[lid] = acc;
+        pcl_ldata[pcl_lid] = pcl_acc;
 
         <%
           cur_size = group_size
         %>
 
-        % while cur_size > no_sync_size:
+        % while cur_size > 1:
             barrier(CLK_LOCAL_MEM_FENCE);
 
             <%
@@ -93,52 +100,18 @@ KERNEL = """//CL//
             assert new_size * 2 == cur_size
             %>
 
-            if (lid < ${new_size})
+            if (pcl_lid < ${new_size})
             {
-                ldata[lid] = REDUCE(
-                  ldata[lid],
-                  ldata[lid + ${new_size}]);
+                pcl_ldata[pcl_lid] = PCL_REDUCE(
+                  pcl_ldata[pcl_lid],
+                  pcl_ldata[pcl_lid + ${new_size}]);
             }
 
             <% cur_size = new_size %>
 
         % endwhile
 
-        % if cur_size > 1:
-            ## we need to synchronize one last time for entry into the
-            ## no-sync region.
-
-            barrier(CLK_LOCAL_MEM_FENCE);
-
-            <%
-            # NB: There's an exact duplicate of this calculation in the
-            # %while loop below.
-
-            new_size = cur_size // 2
-            assert new_size * 2 == cur_size
-            %>
-
-            if (lid < ${new_size})
-            {
-                __local volatile out_type *lvdata = ldata;
-                % while cur_size > 1:
-                    <%
-                    new_size = cur_size // 2
-                    assert new_size * 2 == cur_size
-                    %>
-
-                    lvdata[lid] = REDUCE(
-                      lvdata[lid],
-                      lvdata[lid + ${new_size}]);
-
-                    <% cur_size = new_size %>
-
-                % endwhile
-
-            }
-        % endif
-
-        if (lid == 0) out[get_group_id(0)] = ldata[0];
+        if (pcl_lid == 0) pcl_out[get_group_id(0)] = pcl_ldata[0];
     }
     """
 
@@ -182,32 +155,17 @@ def _get_reduction_source(
 
     # }}}
 
-    # {{{ compute synchronization-less group size
-
-    def get_dev_no_sync_size(device):
-        from pyopencl.characterize import get_simd_group_size
-        result = get_simd_group_size(device, out_type_size)
-
-        if result is None:
-            from warnings import warn
-            warn("Reduction might be unnecessarily slow: "
-                    "can't query SIMD group size")
-            return 1
-
-        return result
-
-    no_sync_size = min(get_dev_no_sync_size(dev) for dev in devices)
-
-    # }}}
-
     from mako.template import Template
-    from pytools import all
     from pyopencl.characterize import has_double_support
+
+    arguments = ", ".join(arg.declarator() for arg in parsed_args)
+    if parsed_args:
+        arguments += ", "
+
     src = str(Template(KERNEL).render(
         out_type=out_type,
-        arguments=", ".join(arg.declarator() for arg in parsed_args),
         group_size=group_size,
-        no_sync_size=no_sync_size,
+        arguments=arguments,
         neutral=neutral,
         reduce_expr=_process_code_for_macro(reduce_expr),
         map_expr=_process_code_for_macro(map_expr),
@@ -232,7 +190,7 @@ def get_reduction_kernel(stage,
          ctx, dtype_out,
          neutral, reduce_expr, map_expr=None, arguments=None,
          name="reduce_kernel", preamble="",
-         device=None, options=[], max_group_size=None):
+         device=None, options=None, max_group_size=None):
 
     if map_expr is None:
         if stage == 2:
@@ -267,9 +225,10 @@ def get_reduction_kernel(stage,
     inf.arg_types = arguments
 
     inf.kernel.set_scalar_arg_dtypes(
-            [None]
+            [None, np.int64]
             + get_arg_list_scalar_arg_dtypes(inf.arg_types)
-            + [np.uint32]*2)
+            + [np.int64]*3 + [np.uint32, np.int64]
+            )
 
     return inf
 
@@ -281,7 +240,7 @@ def get_reduction_kernel(stage,
 class ReductionKernel:
     def __init__(self, ctx, dtype_out,
             neutral, reduce_expr, map_expr=None, arguments=None,
-            name="reduce_kernel", options=[], preamble=""):
+            name="reduce_kernel", options=None, preamble=""):
 
         dtype_out = self.dtype_out = np.dtype(dtype_out)
 
@@ -313,25 +272,43 @@ class ReductionKernel:
                 name=name+"_stage2", options=options, preamble=preamble,
                 max_group_size=max_group_size)
 
-        from pytools import any
-        from pyopencl.tools import VectorArg
-        assert any(
-                isinstance(arg_tp, VectorArg)
-                for arg_tp in self.stage_1_inf.arg_types), \
-                "ReductionKernel can only be used with functions " \
-                "that have at least one vector argument"
-
     def __call__(self, *args, **kwargs):
-        MAX_GROUP_COUNT = 1024
-        SMALL_SEQ_COUNT = 4
+        """
+        :arg range: A :class:`slice` object. Specifies the range of indices on which
+            the kernel will be executed. May not be given at the same time
+            as *slice*.
+        :arg slice: A :class:`slice` object.
+            Specifies the range of indices on which the kernel will be
+            executed, relative to the first vector-like argument.
+            May not be given at the same time as *range*.
+        :arg allocator:
+
+        .. versionchanged:: 2016.2
+
+            *range_* and *slice_* added.
+        """
+        MAX_GROUP_COUNT = 1024  # noqa
+        SMALL_SEQ_COUNT = 4  # noqa
 
         from pyopencl.array import empty
+        empty = args[0].__class__ if args else empty
 
         stage_inf = self.stage_1_inf
 
         queue = kwargs.pop("queue", None)
+        allocator = kwargs.pop("allocator", None)
         wait_for = kwargs.pop("wait_for", None)
         return_event = kwargs.pop("return_event", False)
+        out = kwargs.pop("out", None)
+
+        if wait_for is None:
+            wait_for = []
+        else:
+            # We'll be modifying it below.
+            wait_for = list(wait_for)
+
+        range_ = kwargs.pop("range", None)
+        slice_ = kwargs.pop("slice", None)
 
         if kwargs:
             raise TypeError("invalid keyword argument to reduction kernel")
@@ -353,43 +330,101 @@ class ReductionKernel:
                     invocation_args.append(arg.base_data)
                     if arg_tp.with_offset:
                         invocation_args.append(arg.offset)
+                    wait_for.extend(arg.events)
                 else:
                     invocation_args.append(arg)
 
-            repr_vec = vectors[0]
-            sz = repr_vec.size
+            if vectors:
+                repr_vec = vectors[0]
+            else:
+                repr_vec = None
+
+            # {{{ range/slice processing
+
+            if range_ is not None:
+                if slice_ is not None:
+                    raise TypeError("may not specify both range and slice "
+                            "keyword arguments")
+
+            else:
+                if slice_ is None:
+                    slice_ = slice(None)
+
+                if repr_vec is None:
+                    raise TypeError(
+                            "must have vector argument when range is not specified")
+
+                range_ = slice(*slice_.indices(repr_vec.size))
+
+            assert range_ is not None
+
+            start = range_.start
+            if start is None:
+                start = 0
+            if range_.step is None:
+                step = 1
+            else:
+                step = range_.step
+            sz = abs(range_.stop - start)//step
+
+            # }}}
 
             if queue is not None:
                 use_queue = queue
             else:
+                if repr_vec is None:
+                    raise TypeError(
+                            "must specify queue argument when no vector argument "
+                            "present")
+
                 use_queue = repr_vec.queue
 
-            if sz <= stage_inf.group_size*SMALL_SEQ_COUNT*MAX_GROUP_COUNT:
+            if allocator is None:
+                if repr_vec is None:
+                    from pyopencl.tools import DeferredAllocator
+                    allocator = DeferredAllocator(queue.context)
+                else:
+                    allocator = repr_vec.allocator
+
+            if sz == 0:
+                result = empty(use_queue, (), self.dtype_out, allocator=allocator)
+                group_count = 1
+                seq_count = 0
+
+            elif sz <= stage_inf.group_size*SMALL_SEQ_COUNT*MAX_GROUP_COUNT:
                 total_group_size = SMALL_SEQ_COUNT*stage_inf.group_size
                 group_count = (sz + total_group_size - 1) // total_group_size
                 seq_count = SMALL_SEQ_COUNT
+
             else:
                 group_count = MAX_GROUP_COUNT
                 macrogroup_size = group_count*stage_inf.group_size
                 seq_count = (sz + macrogroup_size - 1) // macrogroup_size
 
-            if group_count == 1:
+            size_args = [start, step, range_.stop, seq_count, sz]
+
+            if group_count == 1 and out is not None:
+                result = out
+            elif group_count == 1:
                 result = empty(use_queue,
                         (), self.dtype_out,
-                        allocator=repr_vec.allocator)
+                        allocator=allocator)
             else:
                 result = empty(use_queue,
                         (group_count,), self.dtype_out,
-                        allocator=repr_vec.allocator)
+                        allocator=allocator)
 
             last_evt = stage_inf.kernel(
                     use_queue,
                     (group_count*stage_inf.group_size,),
                     (stage_inf.group_size,),
-                    *([result.data]+invocation_args+[seq_count, sz]),
-                    **dict(wait_for=wait_for))
+                    *([result.base_data, result.offset]
+                        + invocation_args + size_args),
+                    wait_for=wait_for)
             wait_for = [last_evt]
 
+            result.add_event(last_evt)
+
             if group_count == 1:
                 if return_event:
                     return result, last_evt
@@ -399,6 +434,8 @@ class ReductionKernel:
                 stage_inf = self.stage_2_inf
                 args = (result,) + stage1_args
 
+                range_ = slice_ = None
+
 # }}}
 
 
@@ -407,9 +444,12 @@ class ReductionKernel:
 class ReductionTemplate(KernelTemplateBase):
     def __init__(self,
             arguments, neutral, reduce_expr, map_expr=None,
-            is_segment_start_expr=None, input_fetch_exprs=[],
+            is_segment_start_expr=None, input_fetch_exprs=None,
             name_prefix="reduce", preamble="", template_processor=None):
 
+        if input_fetch_exprs is None:
+            input_fetch_exprs = []
+
         KernelTemplateBase.__init__(
                 self, template_processor=template_processor)
         self.arguments = arguments
@@ -421,7 +461,7 @@ class ReductionTemplate(KernelTemplateBase):
 
     def build_inner(self, context, type_aliases=(), var_values=(),
             more_preamble="", more_arguments=(), declare_types=(),
-            options=(), devices=None):
+            options=None, devices=None):
         renderer = self.get_renderer(
                 type_aliases, var_values, context, options)
 
@@ -435,7 +475,7 @@ class ReductionTemplate(KernelTemplateBase):
                 renderer(self.neutral), renderer(self.reduce_expr),
                 renderer(self.map_expr),
                 renderer.render_argument_list(self.arguments, more_arguments),
-                name=renderer(self.name_prefix), options=list(options),
+                name=renderer(self.name_prefix), options=options,
                 preamble=(
                     type_decl_preamble
                     + "\n"
@@ -467,7 +507,15 @@ def get_sum_kernel(ctx, dtype_out, dtype_in):
     if dtype_out is None:
         dtype_out = dtype_in
 
-    return ReductionKernel(ctx, dtype_out, "0", "a+b",
+    reduce_expr = "a+b"
+    neutral_expr = "0"
+    if dtype_out.kind == "c":
+        from pyopencl.elementwise import complex_dtype_to_name
+        dtname = complex_dtype_to_name(dtype_out)
+        reduce_expr = "%s_add(a, b)" % dtname
+        neutral_expr = "%s_new(0, 0)" % dtname
+
+    return ReductionKernel(ctx, dtype_out, neutral_expr, reduce_expr,
             arguments="const %(tp)s *in"
             % {"tp": dtype_to_ctype(dtype_in)})
 
@@ -486,50 +534,31 @@ def _get_dot_expr(dtype_out, dtype_a, dtype_b, conjugate_first,
                 dtype_a.type(0), dtype_b.type(0),
                 has_double_support)
 
-    a_real_dtype = dtype_a.type(0).real.dtype
-    b_real_dtype = dtype_b.type(0).real.dtype
-    out_real_dtype = dtype_out.type(0).real.dtype
-
     a_is_complex = dtype_a.kind == "c"
     b_is_complex = dtype_b.kind == "c"
-    out_is_complex = dtype_out.kind == "c"
 
     from pyopencl.elementwise import complex_dtype_to_name
 
-    if a_is_complex and b_is_complex:
-        a = "a[%s]" % index_expr
-        b = "b[%s]" % index_expr
-        if dtype_a != dtype_out:
-            a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), a)
-        if dtype_b != dtype_out:
-            b = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), b)
+    a = "a[%s]" % index_expr
+    b = "b[%s]" % index_expr
 
-        if conjugate_first and a_is_complex:
-            a = "%s_conj(%s)" % (
-                    complex_dtype_to_name(dtype_out), a)
+    if a_is_complex and (dtype_a != dtype_out):
+        a = "{}_cast({})".format(complex_dtype_to_name(dtype_out), a)
+    if b_is_complex and (dtype_b != dtype_out):
+        b = "{}_cast({})".format(complex_dtype_to_name(dtype_out), b)
 
-        map_expr = "%s_mul(%s, %s)" % (
-                complex_dtype_to_name(dtype_out), a, b)
+    if a_is_complex and conjugate_first and a_is_complex:
+        a = "{}_conj({})".format(
+                complex_dtype_to_name(dtype_out), a)
+
+    if a_is_complex and not b_is_complex:
+        map_expr = "{}_mulr({}, {})".format(complex_dtype_to_name(dtype_out), a, b)
+    elif not a_is_complex and b_is_complex:
+        map_expr = "{}_rmul({}, {})".format(complex_dtype_to_name(dtype_out), a, b)
+    elif a_is_complex and b_is_complex:
+        map_expr = "{}_mul({}, {})".format(complex_dtype_to_name(dtype_out), a, b)
     else:
-        a = "a[%s]" % index_expr
-        b = "b[%s]" % index_expr
-
-        if out_is_complex:
-            if a_is_complex and dtype_a != dtype_out:
-                a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), a)
-            if b_is_complex and dtype_b != dtype_out:
-                b = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), b)
-
-            if not a_is_complex and a_real_dtype != out_real_dtype:
-                a = "(%s) (%s)" % (dtype_to_ctype(out_real_dtype), a)
-            if not b_is_complex and b_real_dtype != out_real_dtype:
-                b = "(%s) (%s)" % (dtype_to_ctype(out_real_dtype), b)
-
-        if conjugate_first and a_is_complex:
-            a = "%s_conj(%s)" % (
-                    complex_dtype_to_name(dtype_out), a)
-
-        map_expr = "%s*%s" % (a, b)
+        map_expr = f"{a}*{b}"
 
     return map_expr, dtype_out, dtype_b
 
@@ -542,14 +571,22 @@ def get_dot_kernel(ctx, dtype_out, dtype_a=None, dtype_b=None,
             dtype_out, dtype_a, dtype_b, conjugate_first,
             has_double_support=has_double_support(ctx.devices[0]))
 
-    return ReductionKernel(ctx, dtype_out, neutral="0",
-            reduce_expr="a+b", map_expr=map_expr,
-            arguments=
-            "const %(tp_a)s *a, "
-            "const %(tp_b)s *b" % {
-                "tp_a": dtype_to_ctype(dtype_a),
-                "tp_b": dtype_to_ctype(dtype_b),
-                })
+    reduce_expr = "a+b"
+    neutral_expr = "0"
+    if dtype_out.kind == "c":
+        from pyopencl.elementwise import complex_dtype_to_name
+        dtname = complex_dtype_to_name(dtype_out)
+        reduce_expr = "%s_add(a, b)" % dtname
+        neutral_expr = "%s_new(0, 0)" % dtname
+
+    return ReductionKernel(ctx, dtype_out, neutral=neutral_expr,
+            reduce_expr=reduce_expr, map_expr=map_expr,
+            arguments=(
+                "const %(tp_a)s *a, "
+                "const %(tp_b)s *b" % {
+                    "tp_a": dtype_to_ctype(dtype_a),
+                    "tp_b": dtype_to_ctype(dtype_b),
+                    }))
 
 
 @context_dependent_memoize
@@ -564,14 +601,14 @@ def get_subset_dot_kernel(ctx, dtype_out, dtype_subset, dtype_a=None, dtype_b=No
     # important: lookup_tbl must be first--it controls the length
     return ReductionKernel(ctx, dtype_out, neutral="0",
             reduce_expr="a+b", map_expr=map_expr,
-            arguments=
-            "const %(tp_lut)s *lookup_tbl, "
-            "const %(tp_a)s *a, "
-            "const %(tp_b)s *b" % {
-            "tp_lut": dtype_to_ctype(dtype_subset),
-            "tp_a": dtype_to_ctype(dtype_a),
-            "tp_b": dtype_to_ctype(dtype_b),
-            })
+            arguments=(
+                "const %(tp_lut)s *lookup_tbl, "
+                "const %(tp_a)s *a, "
+                "const %(tp_b)s *b" % {
+                    "tp_lut": dtype_to_ctype(dtype_subset),
+                    "tp_a": dtype_to_ctype(dtype_a),
+                    "tp_b": dtype_to_ctype(dtype_b),
+                    }))
 
 
 def get_minmax_neutral(what, dtype):
@@ -603,10 +640,10 @@ def get_minmax_kernel(ctx, what, dtype):
 
     return ReductionKernel(ctx, dtype,
             neutral=get_minmax_neutral(what, dtype),
-            reduce_expr="%(reduce_expr)s" % {"reduce_expr": reduce_expr},
-            arguments="const %(tp)s *in" % {
-                "tp": dtype_to_ctype(dtype),
-                }, preamble="#define MY_INFINITY (1./0)")
+            reduce_expr=f"{reduce_expr}",
+            arguments="const {tp} *in".format(
+                tp=dtype_to_ctype(dtype),
+                ), preamble="#define MY_INFINITY (1./0)")
 
 
 @context_dependent_memoize
@@ -620,14 +657,15 @@ def get_subset_minmax_kernel(ctx, what, dtype, dtype_subset):
 
     return ReductionKernel(ctx, dtype,
             neutral=get_minmax_neutral(what, dtype),
-            reduce_expr="%(reduce_expr)s" % {"reduce_expr": reduce_expr},
+            reduce_expr=f"{reduce_expr}",
             map_expr="in[lookup_tbl[i]]",
-            arguments=
-            "const %(tp_lut)s *lookup_tbl, "
-            "const %(tp)s *in" % {
-            "tp": dtype_to_ctype(dtype),
-            "tp_lut": dtype_to_ctype(dtype_subset),
-            }, preamble="#define MY_INFINITY (1./0)")
+            arguments=(
+                "const %(tp_lut)s *lookup_tbl, "
+                "const %(tp)s *in" % {
+                    "tp": dtype_to_ctype(dtype),
+                    "tp_lut": dtype_to_ctype(dtype_subset),
+                    }),
+            preamble="#define MY_INFINITY (1./0)")
 
 # }}}
 
