@@ -1,6 +1,5 @@
 """Various helpful bits and pieces without much of a common theme."""
 
-from __future__ import division
 
 __copyright__ = "Copyright (C) 2010 Andreas Kloeckner"
 
@@ -28,10 +27,14 @@ OTHER DEALINGS IN THE SOFTWARE.
 """
 
 
+from sys import intern
+
+# Do not add a pyopencl import here: This will add an import cycle.
+
 import numpy as np
-from decorator import decorator
-import pyopencl as cl
 from pytools import memoize, memoize_method
+from pyopencl._cl import bitlog2  # noqa: F401
+from pytools.persistent_dict import KeyBuilder as KeyBuilderBase
 
 import re
 
@@ -41,44 +44,25 @@ from pyopencl.compyte.dtypes import (  # noqa
 
 
 def _register_types():
-    from pyopencl.compyte.dtypes import _fill_dtype_registry
+    from pyopencl.compyte.dtypes import (
+            TYPE_REGISTRY, fill_registry_with_opencl_c_types)
 
-    _fill_dtype_registry(respect_windows=False, include_bool=False)
+    fill_registry_with_opencl_c_types(TYPE_REGISTRY)
 
     get_or_register_dtype("cfloat_t", np.complex64)
     get_or_register_dtype("cdouble_t", np.complex128)
 
-    is_64_bit = tuple.__itemsize__ * 8 == 64
-    if not is_64_bit:
-        get_or_register_dtype(
-                ["unsigned long", "unsigned long int"], np.uint64)
-        get_or_register_dtype(
-                ["signed long", "signed long int", "long int"], np.int64)
 
 _register_types()
 
 
 # {{{ imported names
 
-bitlog2 = cl.bitlog2
-
-PooledBuffer = cl.PooledBuffer
-
-from pyopencl._cl import _tools_DeferredAllocator as DeferredAllocator
 from pyopencl._cl import (  # noqa
-        _tools_ImmediateAllocator as ImmediateAllocator)
-
-
-class CLAllocator(DeferredAllocator):
-    def __init__(self, *args, **kwargs):
-        from warnings import warn
-        warn("pyopencl.tools.CLAllocator is deprecated. "
-                "It will be continue to exist throughout the 2013.x "
-                "versions of PyOpenCL. Use {Deferred,Immediate}Allocator.",
-                DeprecationWarning, 2)
-        DeferredAllocator.__init__(self, *args, **kwargs)
-
-MemoryPool = cl.MemoryPool
+        PooledBuffer as PooledBuffer,
+        _tools_DeferredAllocator as DeferredAllocator,
+        _tools_ImmediateAllocator as ImmediateAllocator,
+        MemoryPool as MemoryPool)
 
 # }}}
 
@@ -88,32 +72,42 @@ MemoryPool = cl.MemoryPool
 _first_arg_dependent_caches = []
 
 
-@decorator
-def first_arg_dependent_memoize(func, cl_object, *args):
-    """Provides memoization for a function. Typically used to cache
-    things that get created inside a :class:`pyopencl.Context`, e.g. programs
-    and kernels. Assumes that the first argument of the decorated function is
-    an OpenCL object that might go away, such as a :class:`pyopencl.Context` or
-    a :class:`pyopencl.CommandQueue`, and based on which we might want to clear
-    the cache.
+def first_arg_dependent_memoize(func):
+    def wrapper(cl_object, *args, **kwargs):
+        """Provides memoization for a function. Typically used to cache
+        things that get created inside a :class:`pyopencl.Context`, e.g. programs
+        and kernels. Assumes that the first argument of the decorated function is
+        an OpenCL object that might go away, such as a :class:`pyopencl.Context` or
+        a :class:`pyopencl.CommandQueue`, and based on which we might want to clear
+        the cache.
 
-    .. versionadded:: 2011.2
-    """
-    try:
-        ctx_dict = func._pyopencl_first_arg_dep_memoize_dic
-    except AttributeError:
-        # FIXME: This may keep contexts alive longer than desired.
-        # But I guess since the memory in them is freed, who cares.
-        ctx_dict = func._pyopencl_first_arg_dep_memoize_dic = {}
-        _first_arg_dependent_caches.append(ctx_dict)
+        .. versionadded:: 2011.2
+        """
+        if kwargs:
+            cache_key = (args, frozenset(kwargs.items()))
+        else:
+            cache_key = (args,)
+
+        try:
+            ctx_dict = func._pyopencl_first_arg_dep_memoize_dic
+        except AttributeError:
+            # FIXME: This may keep contexts alive longer than desired.
+            # But I guess since the memory in them is freed, who cares.
+            ctx_dict = func._pyopencl_first_arg_dep_memoize_dic = {}
+            _first_arg_dependent_caches.append(ctx_dict)
+
+        try:
+            return ctx_dict[cl_object][cache_key]
+        except KeyError:
+            arg_dict = ctx_dict.setdefault(cl_object, {})
+            result = func(cl_object, *args, **kwargs)
+            arg_dict[cache_key] = result
+            return result
+
+    from functools import update_wrapper
+    update_wrapper(wrapper, func)
+    return wrapper
 
-    try:
-        return ctx_dict[cl_object][args]
-    except KeyError:
-        arg_dict = ctx_dict.setdefault(cl_object, {})
-        result = func(cl_object, *args)
-        arg_dict[args] = result
-        return result
 
 context_dependent_memoize = first_arg_dependent_memoize
 
@@ -133,8 +127,8 @@ def first_arg_dependent_memoize_nested(nested_func):
 
     from functools import wraps
     cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
-            % (nested_func.__name__, nested_func.func_code.co_filename,
-                nested_func.func_code.co_firstlineno))
+            % (nested_func.__name__, nested_func.__code__.co_filename,
+                nested_func.__code__.co_firstlineno))
 
     from inspect import currentframe
     # prevent ref cycle
@@ -177,18 +171,46 @@ def clear_first_arg_caches():
     for cache in _first_arg_dependent_caches:
         cache.clear()
 
+
 import atexit
 atexit.register(clear_first_arg_caches)
 
 # }}}
 
 
+# {{{ pytest fixtures
+
+class _ContextFactory:
+    def __init__(self, device):
+        self.device = device
+
+    def __call__(self):
+        # Get rid of leftovers from past tests.
+        # CL implementations are surprisingly limited in how many
+        # simultaneous contexts they allow...
+        clear_first_arg_caches()
+
+        from gc import collect
+        collect()
+
+        import pyopencl as cl
+        return cl.Context([self.device])
+
+    def __str__(self):
+        # Don't show address, so that parallel test collection works
+        return ("<context factory for <pyopencl.Device '%s' on '%s'>>" %
+                (self.device.name.strip(),
+                 self.device.platform.name.strip()))
+
+
 def get_test_platforms_and_devices(plat_dev_string=None):
     """Parse a string of the form 'PYOPENCL_TEST=0:0,1;intel:i5'.
 
     :return: list of tuples (platform, [device, device, ...])
     """
 
+    import pyopencl as cl
+
     if plat_dev_string is None:
         import os
         plat_dev_string = os.environ.get("PYOPENCL_TEST", None)
@@ -203,7 +225,7 @@ def get_test_platforms_and_devices(plat_dev_string=None):
 
         found = False
         for obj in objs:
-            if identifier.lower() in (obj.name + ' ' + obj.vendor).lower():
+            if identifier.lower() in (obj.name + " " + obj.vendor).lower():
                 return obj
         if not found:
             raise RuntimeError("object '%s' not found" % identifier)
@@ -238,65 +260,74 @@ def get_test_platforms_and_devices(plat_dev_string=None):
                 for platform in cl.get_platforms()]
 
 
+def get_pyopencl_fixture_arg_names(metafunc, extra_arg_names=None):
+    if extra_arg_names is None:
+        extra_arg_names = []
+
+    supported_arg_names = [
+            "platform", "device",
+            "ctx_factory", "ctx_getter",
+            ] + extra_arg_names
+
+    arg_names = []
+    for arg in supported_arg_names:
+        if arg not in metafunc.fixturenames:
+            continue
+
+        if arg == "ctx_getter":
+            from warnings import warn
+            warn("The 'ctx_getter' arg is deprecated in "
+                    "favor of 'ctx_factory'.",
+                    DeprecationWarning)
+
+        arg_names.append(arg)
+
+    return arg_names
+
+
+def get_pyopencl_fixture_arg_values():
+    import pyopencl as cl
+
+    arg_values = []
+    for platform, devices in get_test_platforms_and_devices():
+        for device in devices:
+            arg_dict = {
+                "platform": platform,
+                "device": device,
+                "ctx_factory": _ContextFactory(device),
+                "ctx_getter": _ContextFactory(device)
+            }
+            arg_values.append(arg_dict)
+
+    def idfn(val):
+        if isinstance(val, cl.Platform):
+            # Don't show address, so that parallel test collection works
+            return f"<pyopencl.Platform '{val.name}'>"
+        else:
+            return str(val)
+
+    return arg_values, idfn
+
+
 def pytest_generate_tests_for_pyopencl(metafunc):
-    class ContextFactory:
-        def __init__(self, device):
-            self.device = device
+    arg_names = get_pyopencl_fixture_arg_names(metafunc)
+    if not arg_names:
+        return
 
-        def __call__(self):
-            # Get rid of leftovers from past tests.
-            # CL implementations are surprisingly limited in how many
-            # simultaneous contexts they allow...
+    arg_values, ids = get_pyopencl_fixture_arg_values()
+    arg_values = [
+            tuple(arg_dict[name] for name in arg_names)
+            for arg_dict in arg_values
+            ]
 
-            clear_first_arg_caches()
+    metafunc.parametrize(arg_names, arg_values, ids=ids)
 
-            from gc import collect
-            collect()
-
-            return cl.Context([self.device])
-
-        def __str__(self):
-            return "<context factory for %s>" % self.device
-
-    test_plat_and_dev = get_test_platforms_and_devices()
-
-    if ("device" in metafunc.funcargnames
-            or "ctx_factory" in metafunc.funcargnames
-            or "ctx_getter" in metafunc.funcargnames):
-        arg_dict = {}
-
-        for platform, plat_devs in test_plat_and_dev:
-            if "platform" in metafunc.funcargnames:
-                arg_dict["platform"] = platform
-
-            for device in plat_devs:
-                if "device" in metafunc.funcargnames:
-                    arg_dict["device"] = device
-
-                if "ctx_factory" in metafunc.funcargnames:
-                    arg_dict["ctx_factory"] = ContextFactory(device)
-
-                if "ctx_getter" in metafunc.funcargnames:
-                    from warnings import warn
-                    warn("The 'ctx_getter' arg is deprecated in "
-                            "favor of 'ctx_factory'.",
-                            DeprecationWarning)
-                    arg_dict["ctx_getter"] = ContextFactory(device)
-
-                metafunc.addcall(funcargs=arg_dict.copy(),
-                        id=", ".join("%s=%s" % (arg, value)
-                                for arg, value in arg_dict.iteritems()))
-
-    elif "platform" in metafunc.funcargnames:
-        for platform, plat_devs in test_plat_and_dev:
-            metafunc.addcall(
-                    funcargs=dict(platform=platform),
-                    id=str(platform))
+# }}}
 
 
 # {{{ C argument lists
 
-class Argument(object):
+class Argument:
     pass
 
 
@@ -306,31 +337,49 @@ class DtypedArgument(Argument):
         self.name = name
 
     def __repr__(self):
-        return "%s(%r, %s)" % (
+        return "{}({!r}, {})".format(
                 self.__class__.__name__,
                 self.name,
                 self.dtype)
 
+    def __eq__(self, other):
+        return (type(self) == type(other)
+                and self.dtype == other.dtype
+                and self.name == other.name)
+
+    def __hash__(self):
+        return (
+                hash(type(self))
+                ^ hash(self.dtype)
+                ^ hash(self.name))
+
 
 class VectorArg(DtypedArgument):
     def __init__(self, dtype, name, with_offset=False):
-        DtypedArgument.__init__(self, dtype, name)
+        super().__init__(dtype, name)
         self.with_offset = with_offset
 
     def declarator(self):
         if self.with_offset:
             # Two underscores -> less likelihood of a name clash.
-            return "__global %s *%s__base, long %s__offset" % (
+            return "__global {} *{}__base, long {}__offset".format(
                     dtype_to_ctype(self.dtype), self.name, self.name)
         else:
-            result = "__global %s *%s" % (dtype_to_ctype(self.dtype), self.name)
+            result = "__global {} *{}".format(dtype_to_ctype(self.dtype), self.name)
 
         return result
 
+    def __eq__(self, other):
+        return (super().__eq__(other)
+                and self.with_offset == other.with_offset)
+
+    def __hash__(self):
+        return super().__hash__() ^ hash(self.with_offset)
+
 
 class ScalarArg(DtypedArgument):
     def declarator(self):
-        return "%s %s" % (dtype_to_ctype(self.dtype), self.name)
+        return "{} {}".format(dtype_to_ctype(self.dtype), self.name)
 
 
 class OtherArg(Argument):
@@ -341,6 +390,17 @@ class OtherArg(Argument):
     def declarator(self):
         return self.decl
 
+    def __eq__(self, other):
+        return (type(self) == type(other)
+                and self.decl == other.decl
+                and self.name == other.name)
+
+    def __hash__(self):
+        return (
+                hash(type(self))
+                ^ hash(self.decl)
+                ^ hash(self.name))
+
 
 def parse_c_arg(c_arg, with_offset=False):
     for aspace in ["__local", "__constant"]:
@@ -351,8 +411,8 @@ def parse_c_arg(c_arg, with_offset=False):
     c_arg = c_arg.replace("__global", "")
 
     if with_offset:
-        vec_arg_factory = lambda dtype, name: \
-                VectorArg(dtype, name, with_offset=True)
+        def vec_arg_factory(dtype, name):
+            return VectorArg(dtype, name, with_offset=True)
     else:
         vec_arg_factory = VectorArg
 
@@ -379,6 +439,20 @@ def parse_arg_list(arguments, with_offset=False):
     return [parse_single_arg(arg) for arg in arguments]
 
 
+def get_arg_list_arg_types(arg_types):
+    result = []
+
+    for arg_type in arg_types:
+        if isinstance(arg_type, ScalarArg):
+            result.append(arg_type.dtype)
+        elif isinstance(arg_type, VectorArg):
+            result.append(arg_type)
+        else:
+            raise RuntimeError("arg type not understood: %s" % type(arg_type))
+
+    return tuple(result)
+
+
 def get_arg_list_scalar_arg_dtypes(arg_types):
     result = []
 
@@ -409,25 +483,28 @@ def get_arg_offset_adjuster_code(arg_types):
 
     return "\n".join(result)
 
-
 # }}}
 
 
 def get_gl_sharing_context_properties():
+    import pyopencl as cl
+
     ctx_props = cl.context_properties
 
-    from OpenGL import platform as gl_platform, GLX, WGL
+    from OpenGL import platform as gl_platform
 
     props = []
 
     import sys
     if sys.platform in ["linux", "linux2"]:
+        from OpenGL import GLX
         props.append(
-            (ctx_props.GL_CONTEXT_KHR, gl_platform.GetCurrentContext()))
+            (ctx_props.GL_CONTEXT_KHR, GLX.glXGetCurrentContext()))
         props.append(
                 (ctx_props.GLX_DISPLAY_KHR,
                     GLX.glXGetCurrentDisplay()))
     elif sys.platform == "win32":
+        from OpenGL import WGL
         props.append(
             (ctx_props.GL_CONTEXT_KHR, gl_platform.GetCurrentContext()))
         props.append(
@@ -467,11 +544,15 @@ class _CDeclList:
         if dtype in self.declared_dtypes:
             return
 
-        from pyopencl.array import vec
-        if dtype in vec.type_to_scalar_and_count:
+        import pyopencl.cltypes
+        if dtype in pyopencl.cltypes.vec_type_to_scalar_and_count:
             return
 
-        for name, field_data in dtype.fields.iteritems():
+        if hasattr(dtype, "subdtype") and dtype.subdtype is not None:
+            self.add_dtype(dtype.subdtype[0])
+            return
+
+        for _name, field_data in sorted(dtype.fields.items()):
             field_dtype, offset = field_data[:2]
             self.add_dtype(field_dtype)
 
@@ -500,8 +581,12 @@ class _CDeclList:
 
         if self.saw_double:
             result = (
-                    "#pragma OPENCL EXTENSION cl_khr_fp64: enable\n"
-                    "#define PYOPENCL_DEFINE_CDOUBLE\n"
+                    """
+                    #if __OPENCL_C_VERSION__ < 120
+                    #pragma OPENCL EXTENSION cl_khr_fp64: enable
+                    #endif
+                    #define PYOPENCL_DEFINE_CDOUBLE
+                    """
                     + result)
 
         return result
@@ -545,37 +630,57 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     function, not the original one.
     """
 
-    fields = sorted(dtype.fields.iteritems(),
-            key=lambda (name, (dtype, offset)): offset)
+    import pyopencl as cl
+
+    fields = sorted(dtype.fields.items(),
+            key=lambda name_dtype_offset: name_dtype_offset[1][1])
 
     c_fields = []
-    for field_name, (field_dtype, offset) in fields:
-        c_fields.append("  %s %s;" % (dtype_to_ctype(field_dtype), field_name))
+    for field_name, dtype_and_offset in fields:
+        field_dtype, offset = dtype_and_offset[:2]
+        if hasattr(field_dtype, "subdtype") and field_dtype.subdtype is not None:
+            array_dtype = field_dtype.subdtype[0]
+            if hasattr(array_dtype, "subdtype") and array_dtype.subdtype is not None:
+                raise NotImplementedError("nested array dtypes are not supported")
+            array_dims = field_dtype.subdtype[1]
+            dims_str = ""
+            try:
+                for dim in array_dims:
+                    dims_str += "[%d]" % dim
+            except TypeError:
+                dims_str = "[%d]" % array_dims
+            c_fields.append("  {} {}{};".format(
+                dtype_to_ctype(array_dtype), field_name, dims_str)
+            )
+        else:
+            c_fields.append(
+                    "  {} {};".format(dtype_to_ctype(field_dtype), field_name))
 
-    c_decl = "typedef struct {\n%s\n} %s;\n\n" % (
+    c_decl = "typedef struct {{\n{}\n}} {};\n\n".format(
             "\n".join(c_fields),
             name)
 
     cdl = _CDeclList(device)
-    for field_name, (field_dtype, offset) in fields:
+    for _field_name, dtype_and_offset in fields:
+        field_dtype, offset = dtype_and_offset[:2]
         cdl.add_dtype(field_dtype)
 
     pre_decls = cdl.get_declarations()
 
     offset_code = "\n".join(
             "result[%d] = pycl_offsetof(%s, %s);" % (i+1, name, field_name)
-            for i, (field_name, (field_dtype, offset)) in enumerate(fields))
+            for i, (field_name, _) in enumerate(fields))
 
     src = r"""
         #define pycl_offsetof(st, m) \
-                 ((size_t) ((__local char *) &(dummy.m) \
+                 ((uint) ((__local char *) &(dummy.m) \
                  - (__local char *)&dummy ))
 
         %(pre_decls)s
 
         %(my_decl)s
 
-        __kernel void get_size_and_offsets(__global size_t *result)
+        __kernel void get_size_and_offsets(__global uint *result)
         {
             result[0] = sizeof(%(my_type)s);
             __local %(my_type)s dummy;
@@ -596,25 +701,28 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     knl = prg.build(devices=[device]).get_size_and_offsets
 
     import pyopencl.array  # noqa
-    result_buf = cl.array.empty(queue, 1+len(fields), np.uintp)
+    result_buf = cl.array.empty(queue, 1+len(fields), np.uint32)
     knl(queue, (1,), (1,), result_buf.data)
     queue.finish()
     size_and_offsets = result_buf.get()
 
     size = int(size_and_offsets[0])
 
-    from pytools import any
     offsets = size_and_offsets[1:]
     if any(ofs >= size for ofs in offsets):
         # offsets not plausible
 
         if dtype.itemsize == size:
             # If sizes match, use numpy's idea of the offsets.
-            offsets = [offset
-                    for field_name, (field_dtype, offset) in fields]
+            offsets = [dtype_and_offset[1]
+                    for field_name, dtype_and_offset in fields]
         else:
             raise RuntimeError(
-                    "cannot discover struct layout on '%s'" % device)
+                    "OpenCL compiler reported offsetof() past sizeof() "
+                    "for struct layout on '%s'. "
+                    "This makes no sense, and it's usually indicates a "
+                    "compiler bug. "
+                    "Refusing to discover struct layout." % device)
 
     result_buf.data.release()
     del knl
@@ -622,21 +730,34 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     del queue
     del context
 
-    dtype_arg_dict = dict(
-            names=[field_name for field_name, (field_dtype, offset) in fields],
-            formats=[field_dtype
-                for field_name, (field_dtype, offset) in fields],
-            offsets=[int(x) for x in offsets],
-            itemsize=int(size_and_offsets[0]),
-            )
-    dtype = np.dtype(dtype_arg_dict)
-
-    if dtype.itemsize != size_and_offsets[0]:
-        # "Old" versions of numpy (1.6.x?) silently ignore "itemsize". Boo.
-        dtype_arg_dict["names"].append("_pycl_size_fixer")
-        dtype_arg_dict["formats"].append(np.uint8)
-        dtype_arg_dict["offsets"].append(int(size_and_offsets[0])-1)
+    try:
+        dtype_arg_dict = {
+            "names": [field_name
+                      for field_name, (field_dtype, offset) in fields],
+            "formats": [field_dtype
+                        for field_name, (field_dtype, offset) in fields],
+            "offsets": [int(x) for x in offsets],
+            "itemsize": int(size_and_offsets[0]),
+            }
         dtype = np.dtype(dtype_arg_dict)
+        if dtype.itemsize != size_and_offsets[0]:
+            # "Old" versions of numpy (1.6.x?) silently ignore "itemsize". Boo.
+            dtype_arg_dict["names"].append("_pycl_size_fixer")
+            dtype_arg_dict["formats"].append(np.uint8)
+            dtype_arg_dict["offsets"].append(int(size_and_offsets[0])-1)
+            dtype = np.dtype(dtype_arg_dict)
+    except NotImplementedError:
+        def calc_field_type():
+            total_size = 0
+            padding_count = 0
+            for offset, (field_name, (field_dtype, _)) in zip(offsets, fields):
+                if offset > total_size:
+                    padding_count += 1
+                    yield ("__pycl_padding%d" % padding_count,
+                           "V%d" % offset - total_size)
+                yield field_name, field_dtype
+                total_size = field_dtype.itemsize + offset
+        dtype = np.dtype(list(calc_field_type()))
 
     assert dtype.itemsize == size_and_offsets[0]
 
@@ -645,13 +766,21 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
 
 @memoize
 def dtype_to_c_struct(device, dtype):
+    if dtype.fields is None:
+        return ""
+
+    import pyopencl.cltypes
+    if dtype in pyopencl.cltypes.vec_type_to_scalar_and_count:
+        # Vector types are built-in. Don't try to redeclare those.
+        return ""
+
     matched_dtype, c_decl = match_dtype_to_c_struct(
             device, dtype_to_ctype(dtype), dtype)
 
     def dtypes_match():
         result = len(dtype.fields) == len(matched_dtype.fields)
 
-        for name, val in dtype.fields.iteritems():
+        for name, val in dtype.fields.items():
             result = result and matched_dtype.fields[name] == val
 
         return result
@@ -722,9 +851,9 @@ class _ScalarArgPlaceholder(_ArgumentPlaceholder):
     target_class = ScalarArg
 
 
-class _TemplateRenderer(object):
+class _TemplateRenderer:
     def __init__(self, template, type_aliases, var_values, context=None,
-            options=[]):
+            options=None):
         self.template = template
         self.type_aliases = dict(type_aliases)
         self.var_dict = dict(var_values)
@@ -746,6 +875,7 @@ class _TemplateRenderer(object):
         return str(result)
 
     def get_rendered_kernel(self, txt, kernel_name):
+        import pyopencl as cl
         prg = cl.Program(self.context, self(txt)).build(self.options)
 
         kernel_name_prefix = self.var_dict.get("kernel_name_prefix")
@@ -792,8 +922,8 @@ class _TemplateRenderer(object):
                 all_args.extend(arg_list)
 
         if with_offset:
-            vec_arg_factory = lambda typename, name: \
-                    _VectorArgPlaceholder(typename, name, with_offset=True)
+            def vec_arg_factory(typename, name):
+                return _VectorArgPlaceholder(typename, name, with_offset=True)
         else:
             vec_arg_factory = _VectorArgPlaceholder
 
@@ -828,18 +958,18 @@ class _TemplateRenderer(object):
         if arguments is not None:
             cdl.visit_arguments(arguments)
 
-        for tv in self.type_aliases.itervalues():
+        for _, tv in sorted(self.type_aliases.items()):
             cdl.add_dtype(tv)
 
         type_alias_decls = [
-                "typedef %s %s;" % (dtype_to_ctype(val), name)
-                for name, val in self.type_aliases.iteritems()
+                "typedef {} {};".format(dtype_to_ctype(val), name)
+                for name, val in sorted(self.type_aliases.items())
                 ]
 
         return cdl.get_declarations() + "\n" + "\n".join(type_alias_decls)
 
 
-class KernelTemplateBase(object):
+class KernelTemplateBase:
     def __init__(self, template_processor=None):
         self.template_processor = template_processor
 
@@ -873,13 +1003,16 @@ class KernelTemplateBase(object):
             raise RuntimeError(
                     "unknown template processor '%s'" % proc_match.group(1))
 
-    def get_renderer(self, type_aliases, var_values, context=None, options=[]):
+    def get_renderer(self, type_aliases, var_values, context=None, options=None):
         return _TemplateRenderer(self, type_aliases, var_values)
 
+    def build_inner(self, context, *args, **kwargs):
+        raise NotImplementedError
+
     def build(self, context, *args, **kwargs):
         """Provide caching for an :meth:`build_inner`."""
 
-        cache_key = (context, args, tuple(sorted(kwargs.iteritems())))
+        cache_key = (context, args, tuple(sorted(kwargs.items())))
         try:
             return self.build_cache[cache_key]
         except KeyError:
@@ -913,6 +1046,10 @@ class _CLFakeArrayModule:
         from pyopencl.array import empty
         return empty(self.queue, shape, dtype, order=order)
 
+    def hstack(self, arrays):
+        from pyopencl.array import hstack
+        return hstack(arrays, self.queue)
+
 
 def array_module(a):
     if isinstance(a, np.ndarray):
@@ -926,4 +1063,32 @@ def array_module(a):
 
 # }}}
 
+
+def is_spirv(s):
+    spirv_magic = b"\x07\x23\x02\x03"
+    return (
+            isinstance(s, bytes)
+            and (
+                s[:4] == spirv_magic
+                or s[:4] == spirv_magic[::-1]))
+
+
+# {{{ numpy key types builder
+
+class _NumpyTypesKeyBuilder(KeyBuilderBase):
+    def update_for_VectorArg(self, key_hash, key):  # noqa: N802
+        self.rec(key_hash, key.dtype)
+        self.update_for_str(key_hash, key.name)
+        self.rec(key_hash, key.with_offset)
+
+    def update_for_type(self, key_hash, key):
+        if issubclass(key, np.generic):
+            self.update_for_str(key_hash, key.__name__)
+            return
+
+        raise TypeError("unsupported type for persistent hash keying: %s"
+                % type(key))
+
+# }}}
+
 # vim: foldmethod=marker
