@@ -1,6 +1,3 @@
-#! /usr/bin/env python
-
-
 __copyright__ = "Copyright (C) 2013 Andreas Kloeckner"
 
 __license__ = """
@@ -106,7 +103,7 @@ def test_ranged_elwise_kernel(ctx_factory):
     set_to_seven = ElementwiseKernel(context,
             "float *z", "z[i] = 7", "set_to_seven")
 
-    for i, slc in enumerate([
+    for _i, slc in enumerate([
             slice(5, 20000),
             slice(5, 20000, 17),
             slice(3000, 5, -1),
@@ -1104,6 +1101,8 @@ def test_bitonic_argsort(ctx_factory, size, dtype):
             and dtype == np.float64
             and get_pocl_version(dev.platform) < (1, 0)):
         pytest.xfail("Double precision bitonic sort doesn't work on POCL < 1.0")
+    if (dev.platform.name == "Intel(R) OpenCL" and size == 0):
+        pytest.xfail("size-0 arange fails on Intel CL")
 
     if dtype == np.float64 and not has_double_support(dev):
         from pytest import skip
