@@ -853,7 +853,7 @@ def test_nan_arithmetic(ctx_factory):
         shape = (size,)
         a = np.random.randn(*shape).astype(np.float32)
         from random import randrange
-        for i in range(size // 10):
+        for _i in range(size // 10):
             a[randrange(0, size)] = float("nan")
         return a
 
@@ -965,7 +965,7 @@ def test_slice(ctx_factory):
     b = b_gpu.get()
 
     from random import randrange
-    for i in range(20):
+    for _i in range(20):
         start = randrange(ary_len)
         end = randrange(start, ary_len)
 
@@ -974,7 +974,7 @@ def test_slice(ctx_factory):
 
         assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
-    for i in range(20):
+    for _i in range(20):
         start = randrange(ary_len)
         end = randrange(start, ary_len)
 
@@ -983,7 +983,7 @@ def test_slice(ctx_factory):
 
         assert la.norm(a_gpu.get() - a) == 0
 
-    for i in range(20):
+    for _i in range(20):
         start = randrange(ary_len)
         end = randrange(start, ary_len)
 
@@ -1046,6 +1046,10 @@ def test_comparisons(ctx_factory):
 
         assert (res_dev.get() == res).all()
 
+        res2_dev = op(0, res_dev)
+        res2 = op(0, res)
+        assert (res2_dev.get() == res2).all()
+
 
 def test_any_all(ctx_factory):
     context = ctx_factory()
@@ -1171,12 +1175,12 @@ def test_event_management(ctx_factory):
     y = 2**x
     assert len(y.events) == 1
 
-    for i in range(10):
+    for _i in range(10):
         x.fill(0)
 
     assert len(x.events) == 10
 
-    for i in range(1000):
+    for _i in range(1000):
         x.fill(0)
 
     assert len(x.events) < 100
@@ -1351,7 +1355,7 @@ def test_multi_put(ctx_factory):
     ]
 
     out_compare = [np.zeros((10,), np.float32) for i in range(9)]
-    for i, ary in enumerate(out_compare):
+    for _i, ary in enumerate(out_compare):
         ary[idx.get()] = np.arange(0, 6, dtype=np.float32)
 
     cl_array.multi_put(cl_arrays, idx, out=out_arrays)
@@ -1488,6 +1492,9 @@ def test_zero_size_array(ctx_factory, empty_shape):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
+    if queue.device.platform.name == "Intel(R) OpenCL":
+        pytest.xfail("size-0 arrays fail on Intel CL")
+
     a = cl_array.zeros(queue, empty_shape, dtype=np.float32)
     b = cl_array.zeros(queue, empty_shape, dtype=np.float32)
     b.fill(1)
@@ -1516,6 +1523,250 @@ def test_str_without_queue(ctx_factory):
     print(repr(a))
 
 
+@pytest.mark.parametrize("order", ("F", "C"))
+@pytest.mark.parametrize("input_dims", (1, 2, 3))
+def test_stack(ctx_factory, input_dims, order):
+    # Replicates pytato/test/test_codegen.py::test_stack
+    import pyopencl.array as cla
+    cl_ctx = ctx_factory()
+    queue = cl.CommandQueue(cl_ctx)
+
+    shape = (2, 2, 2)[:input_dims]
+    axis = -1 if order == "F" else 0
+
+    from numpy.random import default_rng
+    rng = default_rng()
+    x_in = rng.random(size=shape)
+    y_in = rng.random(size=shape)
+    x_in = x_in if order == "C" else np.asfortranarray(x_in)
+    y_in = y_in if order == "C" else np.asfortranarray(y_in)
+
+    x = cla.to_device(queue, x_in)
+    y = cla.to_device(queue, y_in)
+
+    np.testing.assert_allclose(cla.stack((x, y), axis=axis).get(),
+                                np.stack((x_in, y_in), axis=axis))
+
+
+def test_assign_different_strides(ctx_factory):
+    cl_ctx = ctx_factory()
+    queue = cl.CommandQueue(cl_ctx)
+
+    from pyopencl.clrandom import rand as clrand
+
+    a = clrand(queue, (20, 30), dtype=np.float32)
+    b = cl_array.empty(queue, (20, 30), dtype=np.float32, order="F")
+    with pytest.raises(NotImplementedError):
+        b[:] = a
+
+
+def test_branch_operations_on_pure_scalars():
+    x = np.random.rand()
+    y = np.random.rand()
+    cond = np.random.choice([False, True])
+
+    np.testing.assert_allclose(np.maximum(x, y),
+                               cl_array.maximum(x, y))
+    np.testing.assert_allclose(np.minimum(x, y),
+                               cl_array.minimum(x, y))
+    np.testing.assert_allclose(np.where(cond, x, y),
+                               cl_array.if_positive(cond, x, y))
+
+
+def test_slice_copy(ctx_factory):
+    cl_ctx = ctx_factory()
+    queue = cl.CommandQueue(cl_ctx)
+
+    x = cl.array.to_device(queue, np.random.rand(96, 27))
+    y = x[::8, ::3]
+    with pytest.raises(RuntimeError):
+        y.copy()
+
+
+@pytest.mark.parametrize("order", ("C", "F"))
+def test_ravel(ctx_factory, order):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    x = np.random.randn(10, 4)
+
+    if order == "F":
+        x = np.asfortranarray(x)
+    elif order == "C":
+        pass
+    else:
+        raise AssertionError
+
+    x_cl = cl.array.to_device(cq, x)
+
+    np.testing.assert_allclose(x_cl.ravel(order=order).get(),
+                               x.ravel(order=order))
+
+
+def test_arithmetic_on_non_scalars(ctx_factory):
+    pytest.importorskip("dataclasses")
+
+    from dataclasses import dataclass
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    @dataclass
+    class ArrayContainer:
+        _data: np.ndarray
+
+        def __eq__(self, other):
+            return ArrayContainer(self._data == other)
+
+    with pytest.raises(TypeError):
+        ArrayContainer(np.ones(100)) + cl.array.zeros(cq, (10,), dtype=np.float64)
+
+
+@pytest.mark.parametrize("which", ("add", "sub", "mul", "truediv"))
+def test_arithmetic_with_device_scalars(ctx_factory, which):
+    import operator
+    from numpy.random import default_rng
+
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    rng = default_rng()
+    ndim = rng.integers(1, 5)
+
+    shape = tuple(rng.integers(2, 7) for i in range(ndim))
+
+    x_in = rng.random(shape)
+    x_cl = cl_array.to_device(cq, x_in)
+    idx = tuple(rng.integers(0, dim) for dim in shape)
+
+    op = getattr(operator, which)
+    res_cl = op(x_cl, x_cl[idx])
+    res_np = op(x_in, x_in[idx])
+
+    np.testing.assert_allclose(res_cl.get(), res_np)
+
+
+@pytest.mark.parametrize("then_type", ["array", "host_scalar", "device_scalar"])
+@pytest.mark.parametrize("else_type", ["array", "host_scalar", "device_scalar"])
+def test_if_positive_with_scalars(ctx_factory, then_type, else_type):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    rng = np.random.default_rng()
+    shape = (512,)
+
+    criterion_np = rng.random(shape)
+    criterion_cl = cl_array.to_device(cq, criterion_np)
+
+    def _get_array_or_scalar(rtype, value):
+        if rtype == "array":
+            ary_np = value + np.zeros(shape, dtype=criterion_cl.dtype)
+            ary_cl = value + cl_array.zeros_like(criterion_cl)
+        elif rtype == "host_scalar":
+            ary_np = ary_cl = value
+        elif rtype == "device_scalar":
+            ary_np = value
+            ary_cl = cl_array.to_device(cq, np.array(value))
+        else:
+            raise ValueError(rtype)
+
+        return ary_np, ary_cl
+
+    then_np, then_cl = _get_array_or_scalar(then_type, 0.0)
+    else_np, else_cl = _get_array_or_scalar(else_type, 1.0)
+
+    result_cl = cl_array.if_positive(criterion_cl < 0.5, then_cl, else_cl)
+    result_np = np.where(criterion_np < 0.5, then_np, else_np)
+
+    np.testing.assert_allclose(result_cl.get(), result_np)
+
+
+def test_maximum_minimum_with_scalars(ctx_factory):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    a_np = np.float64(4.0)
+    a_cl = cl_array.to_device(cq, np.array(a_np)).with_queue(None)
+
+    b_np = np.float64(-3.0)
+    b_cl = cl_array.to_device(cq, np.array(b_np)).with_queue(None)
+
+    result = cl_array.maximum(a_np, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), a_np)
+    result = cl_array.maximum(a_cl, b_np, queue=cq)
+    np.testing.assert_allclose(result.get(), a_np)
+    result = cl_array.maximum(a_cl, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), a_np)
+
+    result = cl_array.minimum(a_np, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), b_np)
+    result = cl_array.minimum(a_cl, b_np, queue=cq)
+    np.testing.assert_allclose(result.get(), b_np)
+    result = cl_array.minimum(a_cl, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), b_np)
+
+
+@pytest.mark.parametrize(("reduction", "supports_initial"), [
+    (cl_array.any, False),
+    (cl_array.all, False),
+    (cl_array.sum, True),
+    (cl_array.max, True),
+    (cl_array.min, True),
+    ])
+def test_empty_reductions_vs_numpy(ctx_factory, reduction, supports_initial):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    # {{{ empty
+
+    x_np = np.array([], dtype=np.float64)
+    x_cl = cl_array.to_device(cq, x_np)
+
+    try:
+        ref = getattr(np, reduction.__name__)(x_np)
+    except ValueError:
+        ref = None
+
+    if ref is None:
+        with pytest.raises(ValueError):
+            reduction(x_cl)
+    else:
+        result = reduction(x_cl)
+        if isinstance(result, cl_array.Array):
+            result = result.get()
+
+        np.testing.assert_allclose(result, ref)
+
+    # }}}
+
+    # {{{ empty with initial
+
+    if supports_initial:
+        ref = getattr(np, reduction.__name__)(x_np, initial=5.0)
+        result = reduction(x_cl, initial=5.0)
+        if isinstance(result, cl_array.Array):
+            result = result.get()
+
+        np.testing.assert_allclose(result, ref)
+
+    # }}}
+
+    # {{{ non-empty with initial
+
+    if supports_initial:
+        x_np = np.linspace(-1, 1, 10)
+        x_cl = cl_array.to_device(cq, x_np)
+
+        ref = getattr(np, reduction.__name__)(x_np, initial=5.0)
+        result = reduction(x_cl, initial=5.0).get()
+        np.testing.assert_allclose(result, ref)
+
+        ref = getattr(np, reduction.__name__)(x_np, initial=-5.0)
+        result = reduction(x_cl, initial=-5.0).get()
+        np.testing.assert_allclose(result, ref)
+
+    # }}}
+
+
 if __name__ == "__main__":
     if len(sys.argv) > 1:
         exec(sys.argv[1])
