@@ -194,7 +194,7 @@ def test_get_info(ctx_factory):
     kernel = prg.all_kernels()[0]
     do_test(kernel, cl.kernel_info)
 
-    for i in range(2):  # exercise cache
+    for _i in range(2):  # exercise cache
         for info_name in dir(cl.kernel_work_group_info):
             if not info_name.startswith("_") and info_name != "to_string":
                 try:
@@ -361,13 +361,15 @@ def test_that_python_args_fail(ctx_factory):
     knl = cl.Kernel(prg, "mult")
     try:
         knl(queue, a.shape, None, a_buf, 2, 3)
-        assert False, "PyOpenCL should not accept bare Python types as arguments"
+        raise AssertionError(
+                "PyOpenCL should not accept bare Python types as arguments")
     except cl.LogicError:
         pass
 
     try:
         prg.mult(queue, a.shape, None, a_buf, float(2), 3)
-        assert False, "PyOpenCL should not accept bare Python types as arguments"
+        raise AssertionError(
+                "PyOpenCL should not accept bare Python types as arguments")
     except cl.LogicError:
         pass
 
@@ -554,7 +556,7 @@ def test_mempool(ctx_factory):
     e0 = 12
 
     for e in range(e0-6, e0-4):
-        for i in range(100):
+        for _i in range(100):
             alloc_queue.append(pool.allocate(1 << e))
             if len(alloc_queue) > 10:
                 alloc_queue.pop(0)
@@ -571,8 +573,7 @@ def test_mempool_2(ctx_factory):
 
     pool = MemoryPool(ImmediateAllocator(queue))
 
-    for i in range(2000):
-        s = randrange(1 << 31) >> randrange(32)
+    for s in [randrange(1 << 31) >> randrange(32) for _ in range(2000)] + [2**30]:
         bin_nr = pool.bin_number(s)
         asize = pool.alloc_size(bin_nr)
 
@@ -581,6 +582,20 @@ def test_mempool_2(ctx_factory):
         assert asize < asize*(1+1/8)
 
 
+def test_mempool_32bit_issues():
+    import struct
+    if struct.calcsize("@P") * 8 < 64:
+        pytest.skip("only relevant on 64-bit systems")
+
+    # https://github.com/inducer/pycuda/issues/282
+    from pyopencl._cl import _TestMemoryPool
+    pool = _TestMemoryPool()
+
+    for i in [30, 31, 32, 33, 34]:
+        for offs in range(-5, 5):
+            pool.allocate(2**i + offs)
+
+
 @pytest.mark.parametrize("allocator_cls", [ImmediateAllocator, DeferredAllocator])
 def test_allocator(ctx_factory, allocator_cls):
     context = ctx_factory()
@@ -809,7 +824,7 @@ def test_buffer_get_host_array(ctx_factory):
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.ALLOC_HOST_PTR, size=100)
     try:
         host_buf2 = buf.get_host_array(25, np.float32)
-        assert False, ("MemoryObject.get_host_array should not accept buffer "
+        raise AssertionError("MemoryObject.get_host_array should not accept buffer "
                        "without USE_HOST_PTR")
     except cl.LogicError:
         pass
@@ -818,7 +833,7 @@ def test_buffer_get_host_array(ctx_factory):
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=host_buf)
     try:
         host_buf2 = buf.get_host_array(25, np.float32)
-        assert False, ("MemoryObject.get_host_array should not accept buffer "
+        raise AssertionError("MemoryObject.get_host_array should not accept buffer "
                        "without USE_HOST_PTR")
     except cl.LogicError:
         pass
@@ -966,6 +981,9 @@ def test_spirv(ctx_factory):
         pytest.skip("SPIR-V program creation only available "
                 "in OpenCL 2.1 and higher")
 
+    if not queue.device.il_version:
+        pytest.skip("SPIR-V program creation not supported by device")
+
     n = 50000
 
     a_dev = cl.clrandom.rand(queue, n, np.float32)
@@ -1146,14 +1164,22 @@ def test_compile_link(ctx_factory):
         {
         }
         """).compile()
+    pi_h__prg = cl.Program(ctx, """//CL//
+        inline float get_pi()
+        {
+            return 3.1415f;
+        }
+        """).compile()
     main_prg = cl.Program(ctx, """//CL//
+        #include "pi.h"
+
         void value_sink(float x);
 
         __kernel void experiment()
         {
-            value_sink(3.1415f + get_global_id(0));
+            value_sink(get_pi() + get_global_id(0));
         }
-        """).compile()
+        """).compile(headers=[("pi.h", pi_h__prg)])
     z = cl.link_program(ctx, [vsink_prg, main_prg], devices=ctx.devices)
     z.experiment(queue, (128**2,), (128,))
     queue.finish()
@@ -1183,8 +1209,8 @@ def test_threaded_nanny_events(ctx_factory):
     def create_arrays_thread(n1=10, n2=20):
         ctx = ctx_factory()
         queue = cl.CommandQueue(ctx)
-        for i1 in range(n2):
-            for i in range(n1):
+        for _i1 in range(n2):
+            for _i in range(n1):
                 acl = cl.array.zeros(queue, 10, dtype=np.float32)
                 acl.get()
             # Garbage collection triggers the error
